# Copyright 2015 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
import numpy as np
from .node import Function, NaiveBayes, NaiveBayesMN, MultipleVariables, Centroid
from .model import Model
from .cython_utils import SelectNumbers
import gc


class Inputs(object):
    def __init__(self, base, vars, functions=None):
        self._base = base
        self._vars = vars
        self._unique_individuals = set()
        if functions is None:
            self._funcs = [NaiveBayes, NaiveBayesMN, MultipleVariables, Centroid]
        else:
            self._funcs = functions
        assert len(self._funcs) <= 4
        c = base._classifier
        tag = 'classification' if c else 'regression'
        self._funcs = [x for x in self._funcs if getattr(x, tag)]
        self.functions()
        self._all_variables_index = 0

    def use_all_variables(self):
        return self._base._use_all_vars_input_functions and self._base.nvar > 1 and self._all_variables_index < self._nfunc

    def all_variables(self):
        f = self._func[self._all_variables_index]
        self._all_variables_index += 1
        base = self._base
        unique_individuals = self._unique_individuals
        v = f([x for x in range(base.nvar)],
              ytr=base._ytr, naive_bayes=base.naive_bayes,
              finite=base._finite, mask=base._mask)
        sig = v.signature()
        unique_individuals.add(sig)
        v.height = 0
        if not v.eval(base.X):
            return None
        if not v.isfinite():
            return None
        if not base._bagging_fitness.set_fitness(v):
            return None
        return v

    def functions(self):
        base = self._base
        density = sum([x.hy.density for x in base.X]) / base.nvar
        # func = [x for x in self._funcs if x.nargs > 0]
        func = self._funcs
        if not len(func):
            self._func = None
            self._nfunc = 0
            return
        if density < self._base._min_density:
            func = [x for x in func if x.density_safe]
        self._nfunc = len(func)
        self._func = func

    def function(self):
        if self._nfunc == 1:
            return self._func
        elif self._nfunc == 2:
            if np.random.random() < 0.5:
                return self._func
            return [self._func[1], self._func[0]]
        elif self._nfunc == 3:
            rnd = np.random.random()
            if rnd < 0.3:
                return self._func
            elif rnd < 0.6:
                return [self._func[1], self._func[0], self._func[2]]
            else:
                return [self._func[2], self._func[0], self._func[1]]
        else:
            func = [x for x in self._func]
            np.random.shuffle(func)
            return func

    def input(self):
        base = self._base
        unique_individuals = self._unique_individuals
        vars = self._vars
        if self._nfunc == 0:
            return None
        for _ in range(base._number_tries_feasible_ind):
            args = []
            func = self.function()
            for f in func:
                nargs = f.nargs
                if len(args):
                    vars.pos -= len(args)
                if vars.empty():
                    return None
                args = vars.get(nargs)
                if len(args) < f.min_nargs:
                    return None
                v = f(args,
                      ytr=base._ytr, naive_bayes=base.naive_bayes,
                      finite=base._finite, mask=base._mask)
                sig = v.signature()
                if sig in unique_individuals:
                    continue
                unique_individuals.add(sig)
                v.height = 0
                if not v.eval(base.X):
                    continue
                if not v.isfinite():
                    continue
                if not base._bagging_fitness.set_fitness(v):
                    continue
                return v
        return None


class BasePopulation(object):
    def __init__(self, base=None,
                 tournament_size=2,
                 classifier=True,
                 labels=None,
                 popsize=10000,
                 random_generations=0,
                 es_extra_test=lambda x: True,
                 negative_selection=True):
        self._base = base
        self._p = []
        self._hist = []
        self._bsf = None
        self._estopping = None
        self._tournament_size = tournament_size
        self._index = None
        self._classifier = classifier
        self._es_extra_test = es_extra_test
        self._labels = labels
        self._logger = logging.getLogger('EvoDAG')
        self._previous_estopping = False
        self._current_popsize = 0
        self._popsize = popsize
        self._inds_replace = 0
        self.generation = 1
        self._random_generations = random_generations
        self._density = 0.0
        self._negative_selection = negative_selection

    @property
    def popsize(self):
        return self._current_popsize

    @property
    def population(self):
        "List containing the population"
        return self._p

    @population.setter
    def population(self, a):
        self._current_popsize = len(a)
        self._p = a

    @property
    def density(self):
        return self._density / self.popsize

    def get_density(self, v):
        try:
            return v.hy.density
        except AttributeError:
            return sum([x.density for x in v.hy]) / len(v.hy)

    @property
    def previous_estopping(self):
        """
        Returns whether the last individual set in the population was
        an early stopping individual
        """
        return self._previous_estopping

    @previous_estopping.setter
    def previous_estopping(self, v):
        self._previous_estopping = v

    @property
    def hist(self):
        "List containing all the individuals generated"
        return self._hist

    @property
    def bsf(self):
        "Best so far"
        return self._bsf

    @property
    def estopping(self):
        "Early stopping individual"
        return self._estopping

    @estopping.setter
    def estopping(self, v):
        self.previous_estopping = False
        if v.fitness_vs is None:
            return None
        flag = False
        if self.estopping is None:
            if not self._es_extra_test(v):
                return None
            self._estopping = v
            flag = True
        elif v.fitness_vs > self.estopping.fitness_vs:
            if not self._es_extra_test(v):
                return None
            self._estopping = v
            flag = True
        if flag:
            self.previous_estopping = flag
            self._base._unfeasible_counter = 0
            vfvs = v.fitness_vs
            self._logger.info('(%i) ES: %0.4f %0.4f' % (v.position,
                                                        v.fitness,
                                                        vfvs))

    @bsf.setter
    def bsf(self, v):
        flag = False
        if self.bsf is None:
            self._bsf = v
            flag = True
        elif v.fitness > self.bsf.fitness:
            self._bsf = v
            flag = True
        if flag:
            if v.fitness_vs is None:
                fvs = ""
            else:
                fvs = "%0.4f" % v.fitness_vs
            fts = "%0.4f" % v.fitness
            self._logger.debug('(%(position)s) BSF: %(fts)s %(fvs)s',
                               {'fts': fts, 'fvs': fvs, 'position': v.position})

    def model(self, v=None):
        "Returns the model of node v"
        if v is None:
            v = self.estopping
        hist = self.hist
        trace = self.trace(v)
        ins = None
        if self._base._probability_calibration is not None:
            node = hist[-1]
            node.normalize()
            X = np.array([x.full_array() for x in node.hy]).T
            y = np.array(self._base._y_klass.full_array())
            mask = np.ones(X.shape[0], dtype=np.bool)
            mask[np.array(self._base._mask_ts.index)] = False
            ins = self._base._probability_calibration().fit(X[mask], y[mask])
        if self._classifier:
            nclasses = self._labels.shape[0]
        else:
            nclasses = None
        m = Model(trace, hist, nvar=self._base._nvar,
                  classifier=self._classifier, labels=self._labels,
                  probability_calibration=ins, nclasses=nclasses)
        return m

    def trace(self, n):
        "Restore the position in the history of individual v's nodes"
        trace_map = {}
        self._trace(n, trace_map)
        s = list(trace_map.keys())
        s.sort()
        return s

    def _trace(self, n, trace_map):
        if n.position in trace_map:
            return
        else:
            trace_map[n.position] = 1
        if isinstance(n, Function) and n.height > 0:
            if isinstance(n.variable, list):
                for x in n.variable:
                    self._trace(self.hist[x], trace_map)
            else:
                self._trace(self.hist[n.variable], trace_map)

    def clean(self, v):
        self._density -= self.get_density(v)
        if self.estopping is not None and v == self.estopping:
            return
        v.y = None
        v.hy = None

    def random_selection(self, negative=False):
        return np.random.randint(self.popsize)

    def random(self):
        res = []
        done = {}
        randint = np.random.randint
        popsize = self.popsize
        for _ in range(self._tournament_size):
            k = randint(popsize)
            while k in done:
                k = randint(popsize)
            done[k] = 1
            res.append(k)
        return res

    def tournament(self, negative=False):
        """Tournament selection and when negative is True it performs negative
        tournament selection"""
        if self.generation <= self._random_generations and not negative:
            return self.random_selection()
        if not self._negative_selection and negative:
            return self.random_selection(negative=negative)
        vars = self.random()
        fit = [(k, self.population[x].fitness) for k, x in enumerate(vars)]
        if negative:
            fit = min(fit, key=lambda x: x[1])
        else:
            fit = max(fit, key=lambda x: x[1])
        index = fit[0]
        return vars[index]

    def variable_input(self, used_inputs):
        base = self._base
        for _ in range(base._number_tries_feasible_ind):
            if used_inputs.empty():
                return None
            var = base._random_leaf(used_inputs.get_one())
            if var is not None:
                return var
        return None

    def create_population(self):
        "Create the initial population"
        base = self._base
        if base._share_inputs:
            used_inputs_var = SelectNumbers([x for x in range(base.nvar)])
            used_inputs_naive = used_inputs_var
        if base._pr_variable == 0:
            used_inputs_var = SelectNumbers([])
            used_inputs_naive = SelectNumbers([x for x in range(base.nvar)])
        elif base._pr_variable == 1:
            used_inputs_var = SelectNumbers([x for x in range(base.nvar)])
            used_inputs_naive = SelectNumbers([])
        else:
            used_inputs_var = SelectNumbers([x for x in range(base.nvar)])
            used_inputs_naive = SelectNumbers([x for x in range(base.nvar)])
        nb_input = Inputs(base, used_inputs_naive, functions=base._input_functions)
        while ((base._all_inputs and not base.stopping_criteria_tl()) or
               (self.popsize < base.popsize and
                not base.stopping_criteria())):
            if base._all_inputs and used_inputs_var.empty() and used_inputs_naive.empty():
                base._init_popsize = self.popsize
                break
            if nb_input.use_all_variables():
                v = nb_input.all_variables()
                if v is None:
                    continue
            elif not used_inputs_var.empty() and np.random.random() < base._pr_variable:
                v = self.variable_input(used_inputs_var)
                if v is None:
                    used_inputs_var.pos = used_inputs_var.size
                    continue
            elif not used_inputs_naive.empty():
                v = nb_input.input()
                if not used_inputs_var.empty() and used_inputs_naive.empty():
                    base._pr_variable = 1
                if v is None:
                    used_inputs_naive.pos = used_inputs_naive.size
                    if not used_inputs_var.empty():
                        base._pr_variable = 1
                    continue
            else:
                gen = self.generation
                self.generation = 0
                v = base.random_offspring()
                self.generation = gen
            self.add(v)

    def add(self, v):
        "Add an individual to the population"
        self.population.append(v)
        self._current_popsize += 1
        v.position = len(self._hist)
        self._hist.append(v)
        self.bsf = v
        self.estopping = v
        self._density += self.get_density(v)

    def replace(self, v):
        """Replace an individual selected by negative tournament selection with
        individual v"""
        if self.popsize < self._popsize:
            return self.add(v)
        k = self.tournament(negative=True)
        self.clean(self.population[k])
        self.population[k] = v
        v.position = len(self._hist)
        self._hist.append(v)
        self.bsf = v
        self.estopping = v
        self._inds_replace += 1
        self._density += self.get_density(v)
        if self._inds_replace == self._popsize:
            self._inds_replace = 0
            self.generation += 1
            gc.collect()


class SteadyState(BasePopulation):
    def create_population(self):
        super(SteadyState, self).create_population()
        if self.popsize > self._popsize:
            self.population.sort(key=lambda x: x.fitness, reverse=True)
            [self.clean(x) for x in self.population[self._popsize:]]
            self.population = self.population[:self._popsize]


class Generational(BasePopulation):
    "Generational GP using a steady-state as base"
    def __init__(self, *args, **kwargs):
        self._inner = []
        super(Generational, self).__init__(*args, **kwargs)

    def replace(self, v):
        if self.popsize < self._popsize:
            return self.add(v)
        v.position = len(self._hist)
        self._hist.append(v)
        self.bsf = v
        self.estopping = v
        self._inner.append(v)
        if len(self._inner) == self._popsize:
            [self.clean(x) for x in self.population]
            self.population = self._inner
            self._inner = []
            self.generation += 1
            gc.collect()


class HGenerational(Generational):
    def input(self, vars):
        base = self._base
        function_set = base.function_set
        function_selection = base._function_selection_ins
        function_selection.density = base.population.density
        function_selection.unfeasible_functions.clear()
        for i in range(base._number_tries_feasible_ind):
            if base._function_selection:
                func_index = function_selection.tournament()
            else:
                func_index = function_selection.random_function()
            func = function_set[func_index]
            args = vars.get(func.nargs)
            try:
                if len(args) < func.min_nargs:
                    return None
            except AttributeError:
                if len(args) < func.nargs:
                    return None
            args = [self.population[x].position for x in args]
            f = base._random_offspring(func, args)
            if f is None:
                vars.pos -= len(args)
                function_selection.unfeasible_functions.add(func_index)
                continue
            function_selection[func_index] = f.fitness
            return f
        return None

    def extra_inputs(self):
        base = self._base
        if self.popsize <= self._popsize:
            base._init_popsize = self.popsize
            return
        previous = 0
        end = self.popsize
        nvar = end
        while nvar > self._popsize:
            _ = [x for x in range(previous, end)]
            vars = SelectNumbers(_)
            while True:
                f = self.input(vars)
                if f is not None:
                    self.add(f)
                else:
                    break
            nvar = self.popsize - end
            previous = end
            end = self.popsize
        base._unfeasible_counter = 0
        base._init_popsize = self.popsize
        if self.popsize > self._popsize:
            self.population.sort(key=lambda x: x.fitness, reverse=True)
            [self.clean(x) for x in self.population[self._popsize:]]
            self.population = self.population[:self._popsize]

    def create_population(self):
        base = self._base
        random = np.random.random
        pr_variable = base._pr_variable
        assert pr_variable < 1
        nvar = base.nvar
        _ = [x for x in range(nvar)]
        used_inputs_naive = SelectNumbers(_)
        nb_input = Inputs(base, used_inputs_naive,
                          functions=base._input_functions)
        variable_input = self.variable_input
        add = self.add
        input = nb_input.input
        while not used_inputs_naive.empty():
            if pr_variable > 0 and random() < pr_variable:
                v = variable_input(used_inputs_naive)
            else:
                v = input()
            if v is None:
                used_inputs_naive.pos = used_inputs_naive.size
                continue
            add(v)
        return self.extra_inputs()



# Copyright 2015 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from nose.tools import assert_almost_equals
try:
    from mock import MagicMock
except ImportError:
    from unittest.mock import MagicMock
from EvoDAG.node import Variable
from SparseArray import SparseArray


cl = np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
               0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
               1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 2, 2, 2, 2, 2, 2, 2,
               2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
               2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2,
               2, 2, 2, 2, 2, 2])

X = np.array([[5.1, 3.5, 1.4, 0.2],
              [4.9,  3. ,  1.4,  0.2],
              [4.7,  3.2,  1.3,  0.2],
              [4.6,  3.1,  1.5,  0.2],
              [5. ,  3.6,  1.4,  0.2],
              [5.4,  3.9,  1.7,  0.4],
              [4.6,  3.4,  1.4,  0.3],
              [5. ,  3.4,  1.5,  0.2],
              [4.4,  2.9,  1.4,  0.2],
              [4.9,  3.1,  1.5,  0.1],
              [5.4,  3.7,  1.5,  0.2],
              [4.8,  3.4,  1.6,  0.2],
              [4.8,  3. ,  1.4,  0.1],
              [4.3,  3. ,  1.1,  0.1],
              [5.8,  4. ,  1.2,  0.2],
              [5.7,  4.4,  1.5,  0.4],
              [5.4,  3.9,  1.3,  0.4],
              [5.1,  3.5,  1.4,  0.3],
              [5.7,  3.8,  1.7,  0.3],
              [5.1,  3.8,  1.5,  0.3],
              [5.4,  3.4,  1.7,  0.2],
              [5.1,  3.7,  1.5,  0.4],
              [4.6,  3.6,  1. ,  0.2],
              [5.1,  3.3,  1.7,  0.5],
              [4.8,  3.4,  1.9,  0.2],
              [5. ,  3. ,  1.6,  0.2],
              [5. ,  3.4,  1.6,  0.4],
              [5.2,  3.5,  1.5,  0.2],
              [5.2,  3.4,  1.4,  0.2],
              [4.7,  3.2,  1.6,  0.2],
              [4.8,  3.1,  1.6,  0.2],
              [5.4,  3.4,  1.5,  0.4],
              [5.2,  4.1,  1.5,  0.1],
              [5.5,  4.2,  1.4,  0.2],
              [4.9,  3.1,  1.5,  0.1],
              [5. ,  3.2,  1.2,  0.2],
              [5.5,  3.5,  1.3,  0.2],
              [4.9,  3.1,  1.5,  0.1],
              [4.4,  3. ,  1.3,  0.2],
              [5.1,  3.4,  1.5,  0.2],
              [5. ,  3.5,  1.3,  0.3],
              [4.5,  2.3,  1.3,  0.3],
              [4.4,  3.2,  1.3,  0.2],
              [5. ,  3.5,  1.6,  0.6],
              [5.1,  3.8,  1.9,  0.4],
              [4.8,  3. ,  1.4,  0.3],
              [5.1,  3.8,  1.6,  0.2],
              [4.6,  3.2,  1.4,  0.2],
              [5.3,  3.7,  1.5,  0.2],
              [5. ,  3.3,  1.4,  0.2],
              [7. ,  3.2,  4.7,  1.4],
              [6.4,  3.2,  4.5,  1.5],
              [6.9,  3.1,  4.9,  1.5],
              [5.5,  2.3,  4. ,  1.3],
              [6.5,  2.8,  4.6,  1.5],
              [5.7,  2.8,  4.5,  1.3],
              [6.3,  3.3,  4.7,  1.6],
              [4.9,  2.4,  3.3,  1. ],
              [6.6,  2.9,  4.6,  1.3],
              [5.2,  2.7,  3.9,  1.4],
              [5. ,  2. ,  3.5,  1. ],
              [5.9,  3. ,  4.2,  1.5],
              [6. ,  2.2,  4. ,  1. ],
              [6.1,  2.9,  4.7,  1.4],
              [5.6,  2.9,  3.6,  1.3],
              [6.7,  3.1,  4.4,  1.4],
              [5.6,  3. ,  4.5,  1.5],
              [5.8,  2.7,  4.1,  1. ],
              [6.2,  2.2,  4.5,  1.5],
              [5.6,  2.5,  3.9,  1.1],
              [5.9,  3.2,  4.8,  1.8],
              [6.1,  2.8,  4. ,  1.3],
              [6.3,  2.5,  4.9,  1.5],
              [6.1,  2.8,  4.7,  1.2],
              [6.4,  2.9,  4.3,  1.3],
              [6.6,  3. ,  4.4,  1.4],
              [6.8,  2.8,  4.8,  1.4],
              [6.7,  3. ,  5. ,  1.7],
              [6. ,  2.9,  4.5,  1.5],
              [5.7,  2.6,  3.5,  1. ],
              [5.5,  2.4,  3.8,  1.1],
              [5.5,  2.4,  3.7,  1. ],
              [5.8,  2.7,  3.9,  1.2],
              [6. ,  2.7,  5.1,  1.6],
              [5.4,  3. ,  4.5,  1.5],
              [6. ,  3.4,  4.5,  1.6],
              [6.7,  3.1,  4.7,  1.5],
              [6.3,  2.3,  4.4,  1.3],
              [5.6,  3. ,  4.1,  1.3],
              [5.5,  2.5,  4. ,  1.3],
              [5.5,  2.6,  4.4,  1.2],
              [6.1,  3. ,  4.6,  1.4],
              [5.8,  2.6,  4. ,  1.2],
              [5. ,  2.3,  3.3,  1. ],
              [5.6,  2.7,  4.2,  1.3],
              [5.7,  3. ,  4.2,  1.2],
              [5.7,  2.9,  4.2,  1.3],
              [6.2,  2.9,  4.3,  1.3],
              [5.1,  2.5,  3. ,  1.1],
              [5.7,  2.8,  4.1,  1.3],
              [6.3,  3.3,  6. ,  2.5],
              [5.8,  2.7,  5.1,  1.9],
              [7.1,  3. ,  5.9,  2.1],
              [6.3,  2.9,  5.6,  1.8],
              [6.5,  3. ,  5.8,  2.2],
              [7.6,  3. ,  6.6,  2.1],
              [4.9,  2.5,  4.5,  1.7],
              [7.3,  2.9,  6.3,  1.8],
              [6.7,  2.5,  5.8,  1.8],
              [7.2,  3.6,  6.1,  2.5],
              [6.5,  3.2,  5.1,  2. ],
              [6.4,  2.7,  5.3,  1.9],
              [6.8,  3. ,  5.5,  2.1],
              [5.7,  2.5,  5. ,  2. ],
              [5.8,  2.8,  5.1,  2.4],
              [6.4,  3.2,  5.3,  2.3],
              [6.5,  3. ,  5.5,  1.8],
              [7.7,  3.8,  6.7,  2.2],
              [7.7,  2.6,  6.9,  2.3],
              [6. ,  2.2,  5. ,  1.5],
              [6.9,  3.2,  5.7,  2.3],
              [5.6,  2.8,  4.9,  2. ],
              [7.7,  2.8,  6.7,  2. ],
              [6.3,  2.7,  4.9,  1.8],
              [6.7,  3.3,  5.7,  2.1],
              [7.2,  3.2,  6. ,  1.8],
              [6.2,  2.8,  4.8,  1.8],
              [6.1,  3. ,  4.9,  1.8],
              [6.4,  2.8,  5.6,  2.1],
              [7.2,  3. ,  5.8,  1.6],
              [7.4,  2.8,  6.1,  1.9],
              [7.9,  3.8,  6.4,  2. ],
              [6.4,  2.8,  5.6,  2.2],
              [6.3,  2.8,  5.1,  1.5],
              [6.1,  2.6,  5.6,  1.4],
              [7.7,  3. ,  6.1,  2.3],
              [6.3,  3.4,  5.6,  2.4],
              [6.4,  3.1,  5.5,  1.8],
              [6. ,  3. ,  4.8,  1.8],
              [6.9,  3.1,  5.4,  2.1],
              [6.7,  3.1,  5.6,  2.4],
              [6.9,  3.1,  5.1,  2.3],
              [5.8,  2.7,  5.1,  1.9],
              [6.8,  3.2,  5.9,  2.3],
              [6.7,  3.3,  5.7,  2.5],
              [6.7,  3. ,  5.2,  2.3],
              [6.3,  2.5,  5. ,  1.9],
              [6.5,  3. ,  5.2,  2. ],
              [6.2,  3.4,  5.4,  2.3],
              [5.9,  3. ,  5.1,  1.8]])


def tonparray(a):
    return np.array(a.full_array())


def test_features():
    from EvoDAG import RootGP
    gp = RootGP(generations=1)
    gp.X = X
    assert gp.nvar == 4
    print(gp.X)
    assert isinstance(gp.X[0], Variable)
    assert isinstance(gp.X[0].hy, SparseArray)
    gp.Xtest = X
    assert gp.X[0].hy.SSE(gp.X[0].hy_test) == 0


def test_create_population():
    from EvoDAG import EvoDAG
    gp = EvoDAG(generations=1, classifier=False, pr_variable=1,
                popsize=4)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    gp.create_population()
    assert_almost_equals(gp.population.popsize, gp.popsize)
    # a = map(lambda (x, y): x == y, zip(gp.population.population,
    #                                    gp.population._hist))
    a = [x == y1 for x, y1 in zip(gp.population.population,
                                  gp.population._hist)]
    assert np.all(a)
    l = [x.variable for x in gp.population.population]
    l.sort()
    # assert np.all(map(lambda (x, y): x == y, enumerate(l)))
    assert np.all([x == y1 for x, y1 in enumerate(l)])


def test_create_population2():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Function, Centroid
    Centroid.nargs = 0
    gp = EvoDAG(generations=1, classifier=False, pr_variable=1,
                popsize=10)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    gp.create_population()
    for i in gp.population.population[4:]:
        assert isinstance(i, Function)
    Centroid.nargs = 2


def test_create_population_cl():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Function, Variable, NaiveBayes, NaiveBayesMN
    gp = EvoDAG(generations=1, popsize=50, multiple_outputs=True)
    gp.X = X
    gp.nclasses(cl)
    gp.y = cl.copy()
    gp.create_population()
    flag = False
    for i in gp.population.population:
        assert isinstance(i, Function) or isinstance(i, Variable)
        if isinstance(i, Function):
            if not (isinstance(i, NaiveBayes) or isinstance(i, NaiveBayesMN)):
                flag = True
    assert flag


def test_best_so_far():
    from EvoDAG import RootGP
    gp = RootGP(generations=1, classifier=False, popsize=4)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    # randint = np.random.randint
    # mock = MagicMock()
    # mock.side_effect = list(range(gp.popsize))
    # np.random.randint = mock
    gp.create_population()
    p = gp.population.population
    index = np.argsort([x.fitness for x in p])[-1]
    print(p[index].fitness, gp.population.bsf.fitness)
    assert gp.population.bsf == p[index]
    # np.random.randint = randint


def test_early_stopping():
    from EvoDAG import RootGP
    gp = RootGP(generations=1, popsize=4, classifier=False)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    randint = np.random.randint
    mock = MagicMock()
    mock.side_effect = list(range(gp.popsize))
    np.random.randint = mock
    gp.create_population()
    p = gp.population.population
    fit = np.array([x.fitness_vs for x in p])
    best = fit.max()
    index = np.where(best == fit)[0][0]
    assert gp.population.estopping == p[index]
    np.random.randint = randint


def test_variable():
    from EvoDAG import RootGP
    gp = RootGP(generations=1, classifier=False,
                popsize=4)
    gp.X = X
    Xtest = [x for x in X]
    Xtest[0] = Xtest[0] + np.inf
    gp.Xtest = Xtest
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    randint = np.random.randint
    mock = MagicMock()
    mock.side_effect = list(range(gp.popsize))
    np.random.randint = mock
    var = gp.random_leaf()
    assert var.isfinite()
    assert var.hy.isfinite()
    np.random.randint = randint


def test_random_leaf():
    from EvoDAG import RootGP
    gp = RootGP(generations=1, classifier=False, popsize=4)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    randint = np.random.randint
    mock = MagicMock(return_value=0)
    np.random.randint = mock
    mask = tonparray(gp._mask).astype(np.bool)
    weight = np.linalg.lstsq(X[mask, 0][:, np.newaxis], y[mask])[0][0]
    var = gp.random_leaf()
    assert isinstance(var, Variable)
    print(weight, var.weight)
    assert_almost_equals(weight, var.weight)
    np.random.randint = randint


def test_random_leaf_inf():
    from EvoDAG import RootGP
    gp = RootGP(generations=1, classifier=False, popsize=4, tr_fraction=1)
    Xc = [x for x in X]
    Xc[0] = Xc[0] + np.inf
    gp.X = Xc
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    randint = np.random.randint
    mock = MagicMock()
    mock.side_effect = list(range(2))
    np.random.randint = mock
    gp.random_leaf()
    try:
        assert mock()
        assert False
    except Exception:
        pass
    np.random.randint = randint


def test_classification_y():
    from EvoDAG import RootGP
    gp = RootGP(generations=1, multiple_outputs=True, popsize=4)
    assert gp._classifier
    gp.X = X
    y = cl.copy()
    gp.nclasses(y)
    gp.y = y
    print(gp._ytr, gp.y)
    for a, b in zip(gp._ytr, gp.y):
        assert a.SSE(b) > 0
        assert a.sum() == 0
        assert b.sum() < 0


def test_regression_y():
    from EvoDAG import RootGP
    gp = RootGP(generations=1, popsize=4, classifier=False)
    assert not gp._classifier
    gp.X = X
    gp.y = cl.copy()
    assert gp._ytr.SSE(gp.y) > 0
    gp = RootGP(generations=1, popsize=4, classifier=False, tr_fraction=1.0)
    gp.X = X
    gp.y = cl.copy()
    assert gp._ytr.SSE(gp.y) == 0


def test_fitness():
    from EvoDAG import RootGP
    gp = RootGP(generations=1, classifier=False,
                popsize=4)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    l = gp.random_leaf()
    assert l.fitness < 0


def test_mask_vs():
    from EvoDAG.node import Add
    from EvoDAG import RootGP
    from EvoDAG.utils import BER
    gp = RootGP(generations=1, popsize=4, multiple_outputs=True)
    assert gp._classifier
    gp.X = X
    y = cl.copy()
    gp.nclasses(y)
    gp.y = y
    m = np.sign(tonparray(gp._mask_vs)).astype(np.bool)
    v = gp.random_leaf()
    v1 = gp.random_leaf()
    v1 = gp.random_leaf()
    a = Add([0, 1], ytr=gp._ytr, mask=gp._mask)
    a.eval([v, v1])
    hy = SparseArray.argmax(a.hy)
    b = BER(y[m], tonparray(hy)[m])
    gp._bagging_fitness.set_fitness(a)
    print(b, a.fitness_vs * 100)
    assert_almost_equals(b, -a.fitness_vs * 100)


def test_BER():
    from EvoDAG.node import Add
    from EvoDAG import RootGP
    from EvoDAG.utils import BER
    gp = RootGP(generations=1, popsize=4, multiple_outputs=True)
    assert gp._classifier
    gp.X = X
    y = cl.copy()
    gp.nclasses(y)
    gp.y = y
    m = np.sign(tonparray(gp._mask_ts)).astype(np.bool)
    v = gp.random_leaf()
    v1 = gp.random_leaf()
    v1 = gp.random_leaf()
    a = Add([0, 1], ytr=gp._ytr, mask=gp._mask)
    a.eval([v, v1])
    hy = SparseArray.argmax(a.hy)
    b = BER(y[m], tonparray(hy)[m])
    gp._bagging_fitness.fitness(a)
    print(b, a.fitness * 100)
    assert_almost_equals(b, -a.fitness * 100)


def test_tournament():
    from EvoDAG import RootGP
    gp = RootGP(generations=1,
                tournament_size=4,
                classifier=False,
                popsize=4)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    randint = np.random.randint
    mock = MagicMock()
    mock.side_effect = list(range(gp.popsize))
    np.random.randint = mock
    gp.create_population()
    j = gp.population.tournament()
    index = np.argsort([x.fitness for x in gp.population.population])[-1]
    assert j == index
    np.random.randint = randint


def test_tournament_negative():
    from EvoDAG import RootGP
    gp = RootGP(generations=1,
                tournament_size=4,
                classifier=False,
                popsize=4)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    randint = np.random.randint
    mock = MagicMock()
    mock.side_effect = list(range(gp.popsize))
    np.random.randint = mock
    gp.create_population()
    j = gp.population.tournament(negative=True)
    index = np.argsort([x.fitness for x in gp.population.population])[0]
    assert j == index
    np.random.randint = randint


def test_random_offspring():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add, Sin
    gp = EvoDAG(generations=1,
                function_set=[Add, Sin],
                multiple_outputs=True,
                seed=1,
                tournament_size=2,
                popsize=10)
    gp.X = X
    gp.nclasses(cl)
    gp.y = cl.copy()
    gp.create_population()
    a = gp.random_offspring()
    assert isinstance(a, Add) or isinstance(a, Sin)
    assert np.isfinite(a.fitness)


def test_replace_individual():
    from EvoDAG import RootGP
    gp = RootGP(generations=1,
                tournament_size=2,
                classifier=False,
                popsize=5)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    gp.create_population()
    print(gp.population.popsize)
    a = gp.random_offspring()
    assert a.position == 0
    gp.population.replace(a)
    assert np.any([x == a for x in gp.population.population])
    print(a.position, len(gp.population.population))
    assert a.position == len(gp.population.population)


def test_X_sparse():
    from EvoDAG import RootGP
    gp = RootGP(generations=1,
                tournament_size=2,
                popsize=10)
    X1 = list(map(SparseArray.fromlist, X.T))
    gp.X = X1


def test_fit_stopping_criteria_gens():
    from EvoDAG import RootGP
    from EvoDAG.node import Add
    Add.nargs = 2
    function_set = [x for x in RootGP()._function_set if x.regression]
    gp = RootGP(generations=2,
                early_stopping_rounds=None,
                tournament_size=2,
                function_set=function_set,
                classifier=False,
                seed=1,
                popsize=4)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    gp.create_population()
    for i in range(gp.popsize):
        assert not gp.stopping_criteria()
        a = gp.random_offspring()
        gp.population.replace(a)
    assert gp.stopping_criteria()


def test_fit_stopping_criteria_estopping():
    from EvoDAG import RootGP
    function_set = [x for x in RootGP()._function_set if x.regression]
    gp = RootGP(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=4,
                function_set=function_set,
                classifier=False,
                seed=0,
                popsize=4)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    gp.create_population()
    while not gp.stopping_criteria():
        a = gp.random_offspring()
        gp.population.replace(a)
    print(len(gp.population.hist) - gp.population.estopping.position)
    assert (len(gp.population.hist) - gp.population.estopping.position) <= 9


def test_fit():
    from EvoDAG import RootGP
    function_set = [x for x in RootGP()._function_set if x.regression]
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp = RootGP(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=-1,
                function_set=function_set,
                classifier=False,
                seed=0,
                popsize=4).fit(X, y, test_set=X)
    assert np.isfinite(gp.population.estopping.fitness)
    assert np.isfinite(gp.population.estopping.fitness_vs)
    assert gp.population.estopping.hy.isfinite()
    assert len(gp.population.hist) > 4


def test_logging():
    from EvoDAG.model import EvoDAG as evodag
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    evodag(generations=np.inf,
           tournament_size=2,
           early_stopping_rounds=-1,
           classifier=False,
           seed=0,
           popsize=10).fit(X, y, test_set=X)


def test_infite_evolution():
    from EvoDAG import RootGP
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    try:
        RootGP(generations=np.inf,
               tournament_size=2,
               tr_fraction=1,
               early_stopping_rounds=-1,
               seed=0,
               popsize=10).fit(X, y, test_set=X)
        assert False
    except RuntimeError:
        pass


def test_predict():
    from EvoDAG import RootGP
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    function_set = [x for x in RootGP()._function_set if x.regression and x.nargs]
    gp = RootGP(generations=np.inf,
                tournament_size=2,
                function_set=function_set,
                early_stopping_rounds=-1,
                classifier=False,
                remove_raw_inputs=False,
                seed=0,
                popsize=10).fit(X[:-10], y[:-10], test_set=X[-10:])
    es = gp.population.estopping
    assert gp.decision_function().SSE(es.hy_test) == 0
    hy_test = es.hy_test
    assert gp.decision_function(X=X[-10:]).SSE(hy_test) == 0
    hy = gp.decision_function(X=X[-10:])
    _ = gp.predict(X=X[-10:])
    assert SparseArray.fromlist(_).SSE(hy) == 0
    assert len(gp.Xtest)


def test_trace():
    from EvoDAG import RootGP
    from EvoDAG.node import Add
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    Add.nargs = 3
    gp = RootGP(generations=np.inf,
                tournament_size=2,
                function_set=[Add],
                classifier=False,
                early_stopping_rounds=-1,
                seed=0,
                popsize=4)
    gp.X = X[:-10]
    gp.Xtest = X[-10:]
    gp.y = y[:-10]
    gp.create_population()
    a = gp.random_offspring()
    gp.population.replace(a)
    print(a.position, a.variable, a._weight, gp.population.hist[0].variable)
    s = gp.trace(a)
    print(len(s), s)
    assert a._weight.shape[0] + 1 == len(s)


def test_class_values():
    from EvoDAG import RootGP
    y = cl.copy()
    mask = y == 0
    y[mask] = 0
    y[~mask] = -1
    gp = RootGP(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=-1,
                classifier=True,
                multiple_outputs=True,
                seed=0,
                popsize=10).fit(X[:-10], y[:-10], test_set=X[-10:])
    assert np.all(gp._labels == np.array([-1, 0]))


# def test_multiclass():
#     from EvoDAG import EvoDAG
#     y = cl.copy()
#     ncl = np.unique(y).shape[0]
#     gp = EvoDAG(generations=np.inf,
#                 tournament_size=2,
#                 early_stopping_rounds=-1,
#                 seed=0,
#                 popsize=100).fit(X, y)
#     assert len(gp._multiclass_instances) == ncl
#     assert gp._multiclass


# def test_multiclass_decision_function():
#     from EvoDAG import EvoDAG
#     y = cl.copy()
#     gp = EvoDAG(generations=np.inf,
#                 tournament_size=2,
#                 early_stopping_rounds=-1,
#                 seed=0,
#                 popsize=100).fit(X, y, test_set=X)
#     d = gp.decision_function()
#     assert len(d) == np.unique(y).shape[0]


# def test_multiclass_predict():
#     from EvoDAG import RootGP
#     y = cl.copy()
#     y[y == 0] = 3
#     gp = RootGP(generations=np.inf,
#                 tournament_size=2,
#                 early_stopping_rounds=-1,
#                 seed=0,
#                 popsize=100).fit(X, y, test_set=X)
#     d = gp.predict()
#     assert np.unique(d).shape[0] == np.unique(y).shape[0]
#     assert np.all(np.unique(d) == np.unique(y))


def test_get_params():
    from EvoDAG import RootGP
    # y = cl.copy()
    gp = RootGP(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=-1,
                seed=0,
                popsize=100)
    p = gp.get_params()
    assert p['generations'] == np.inf
    assert p['popsize'] == 100
    assert p['seed'] == 0


def test_get_clone():
    from EvoDAG import RootGP
    # y = cl.copy()
    gp = RootGP(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=-1,
                seed=0,
                popsize=100)
    gp1 = gp.clone()
    print(gp.popsize, gp1.popsize)
    assert gp.popsize == gp1.popsize
    assert gp._generations == gp1._generations
    assert gp._seed == gp1._seed


def test_labels():
    from EvoDAG import RootGP
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    gp = RootGP(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=-1,
                multiple_outputs=True,
                seed=0,
                popsize=100).fit(X, y, test_set=X)
    m = gp.model()
    hy = m.predict(X=X)
    print(np.unique(hy), np.unique(y))
    print(np.array([1, 2]))
    for k in np.unique(hy):
        assert k in [1, 2]
    # assert np.all(np.unique(hy) == np.array([1, 2]))


def test_height():
    from EvoDAG.node import Mul, NaiveBayesMN, NaiveBayes
    from test_command_line import default_nargs
    from EvoDAG import EvoDAG
    # NaiveBayesMN.nargs = 0
    # NaiveBayes.nargs = 0
    gp = EvoDAG(generations=1,
                seed=1,
                multiple_outputs=True,
                tournament_size=2,
                use_all_vars_input_functions=True,
                input_functions=[NaiveBayes],
                popsize=5)
    gp.X = X
    gp.nclasses(cl)
    gp.y = cl.copy()
    gp.create_population()
    print(NaiveBayes.nargs, NaiveBayesMN.nargs)
    print([(x, x.height) for x in gp.population.population])
    assert np.all([x.height == 0 for x in gp.population.population])
    args = [3, 4]
    f = gp._random_offspring(Mul, args)
    assert f.height == 1
    default_nargs()


def test_regression():
    from EvoDAG.model import EvoDAG as evodag
    x = np.linspace(-1, 1, 100)
    y = 4.3*x**2 + 3.2 * x - 3.2
    gp = evodag(classifier=False,
                popsize=10,
                generations=2).fit([SparseArray.fromlist(x)], y,
                                   test_set=[SparseArray.fromlist(x)])
    model = gp.model
    yh = gp.predict([SparseArray.fromlist(x)])
    assert not model._classifier
    yh1 = model.predict(X=[SparseArray.fromlist(x)])
    spf = SparseArray.fromlist
    assert spf(yh).SSE(spf(yh1)) == 0


def test_unique():
    from EvoDAG import RootGP
    mock = MagicMock(side_effect=RuntimeError('Mock'))
    ui = RootGP.unique_individual
    RootGP.unique_individual = mock
    gp = RootGP(generations=np.inf,
                tournament_size=2,
                unique_individuals=True,
                multiple_outputs=True,
                early_stopping_rounds=-1,
                seed=0,
                popsize=100)
    try:
        gp.fit(X, cl)
    except RuntimeError:
        pass
    RootGP.unique_individual = ui


def test_RSE():
    from EvoDAG import RootGP
    from EvoDAG.utils import RSE as rse
    from EvoDAG.node import Centroid
    Centroid.nargs = 0
    x = np.linspace(-1, 1, 100)
    y = 4.3*x**2 + 3.2 * x - 3.2
    y[10:12] = 0
    gp = RootGP(classifier=False,
                popsize=10,
                generations=2).fit([SparseArray.fromlist(x)], y,
                                   test_set=[SparseArray.fromlist(x)])
    model = gp.model()
    yh = gp.predict()
    assert not model._classifier
    model.predict(X=[SparseArray.fromlist(x)])
    gp._mask = SparseArray.fromlist([2] * yh.shape[0])
    gp._bagging_fitness.fitness_vs(model._hist[-1])
    print(rse(y, yh), model._hist[-1].fitness_vs)
    assert_almost_equals(rse(y, yh),
                         -model._hist[-1].fitness_vs)
    Centroid.nargs = 2


def test_RSE_avg_zero():
    from EvoDAG.bagging_fitness import BaggingFitness
    from EvoDAG.node import Centroid
    from EvoDAG import EvoDAG
    Centroid.nargs = 0
    class B(BaggingFitness):
        def __init__(self, **kw):
            super(B, self).__init__(**kw)
            self._base._times = 0

        def set_regression_mask(self, v):
            base = self._base
            mask = np.ones(v.size())
            if base._times == 0:
                mask[10:12] = 0
            else:
                mask[10:13] = 0
            base._mask = SparseArray.fromlist(mask)
            base._times += 1

    x = np.linspace(-1, 1, 100)
    y = 4.3*x**2 + 3.2 * x - 3.2
    y[10:12] = 0
    gp = EvoDAG(classifier=False,
                popsize=10,
                generations=2)
    gp._bagging_fitness = B(base=gp)
    gp.X = [SparseArray.fromlist(x)]
    gp.y = y
    print(gp._times)
    assert gp._times == 2
    gp.create_population()
    while not gp.stopping_criteria():
        a = gp.random_offspring()
        gp.replace(a)
    Centroid.nargs = 2


def test_population_as_parameter():
    class Mock(object):
        def __init__(self, *args, **kwargs):
            raise RuntimeError('Mock')
    from EvoDAG import RootGP
    mock = Mock
    gp = RootGP(generations=np.inf,
                population_class=mock,
                tournament_size=2,
                multiple_outputs=True,
                unique_individuals=True,
                early_stopping_rounds=-1,
                seed=0,
                popsize=100)
    try:
        gp.fit(X, cl)
        assert False
    except RuntimeError:
        pass


def test_es_extra_test():
    from EvoDAG import RootGP

    class ParticularException(Exception):
        pass

    x = np.linspace(-1, 1, 100)
    y = 4.3*x**2 + 3.2 * x - 3.2
    es_extra_test = RootGP.es_extra_test
    RootGP.es_extra_test = MagicMock(side_effect=ParticularException('Mock'))
    try:
        RootGP(classifier=False,
               popsize=10,
               generations=2).fit([SparseArray.fromlist(x)], y,
                                  test_set=[SparseArray.fromlist(x)])
        assert False
    except ParticularException:
        RootGP.es_extra_test = es_extra_test


def test_fname():
    from EvoDAG.node import Add, Centroid
    from EvoDAG import RootGP
    Centroid.nargs = 0
    x = np.linspace(-1, 1, 100)
    y = 4.3*x**2 + 3.2 * x - 3.2
    Add.nargs = 10
    gp = RootGP(classifier=False,
                popsize=10,
                generations=2).fit([SparseArray.fromlist(x)], y,
                                   test_set=[SparseArray.fromlist(x)])
    assert gp.signature.count('Ad10') == 1
    Centroid.nargs = 2


def test_unfeasible_counter():
    from EvoDAG import RGP
    gp = RGP(generations=np.inf,
             tournament_size=2,
             early_stopping_rounds=-1,
             classifier=False,
             seed=0,
             popsize=100)
    assert gp._unfeasible_counter == 0
    assert gp.unfeasible_offspring() is None
    assert gp._unfeasible_counter == 1


def test_replace_population_previous_estopping():
    from EvoDAG import RGP
    gp = RGP(generations=np.inf,
             tournament_size=2,
             early_stopping_rounds=-1,
             classifier=False,
             seed=0,
             popsize=3)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    gp.create_population()
    gp.unfeasible_offspring()
    es = gp.population.estopping
    for i in range(20):
        n = gp.random_offspring()
        if n.fitness_vs > es.fitness_vs:
            break
    print(es.fitness_vs, n.fitness_vs, gp._unfeasible_counter)
    assert gp._unfeasible_counter >= 1
    gp.replace(n)
    assert gp.population.previous_estopping


def test_add():
    from EvoDAG import RGP
    gp = RGP(generations=np.inf,
             tournament_size=2,
             early_stopping_rounds=-1,
             classifier=False,
             seed=0,
             popsize=3)
    gp.X = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    gp.create_population()
    gp.unfeasible_offspring()
    es = gp.population.estopping
    for i in range(20):
        n = gp.random_offspring()
        if n.fitness_vs > es.fitness_vs:
            break
    gp.population.add(n)
    print(gp._unfeasible_counter)
    assert gp._unfeasible_counter == 0


def test_unfeasible_counter_fit():
    from EvoDAG import RGP

    class RGP2(RGP):
        def replace(self, a):
            self.population.replace(a)
            self._unfeasible_counter = 100

        def add(self, a):
            self.population.add(a)
            self._unfeasible_counter = 100

    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp = RGP2(generations=10,
              tournament_size=2,
              early_stopping_rounds=1,
              classifier=False,
              seed=0,
              popsize=3)
    [gp.unfeasible_offspring() for _ in range(100)]
    assert gp._unfeasible_counter == 100
    gp.fit(X, y)
    print('Hist', len(gp.population.hist), gp.population.hist)
    assert len(gp.population.hist) <= 4


def test_two_instances():
    from EvoDAG import EvoDAG
    y = cl.copy()
    y[:-2] = -1
    y[-2:] = 1
    function_set = [x for x in EvoDAG()._function_set if x.regression and x.nargs]
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                classifier=False,
                function_set=function_set,
                early_stopping_rounds=-1,
                seed=0,
                popsize=10).fit(X, y, test_set=X)
    assert gp


def test_time_limit():
    from EvoDAG import EvoDAG
    import time
    y = cl.copy()
    t = time.time()
    gp = EvoDAG.init(generations=np.inf,
                     tournament_size=2,
                     early_stopping_rounds=100,
                     multiple_outputs=True,
                     time_limit=0.9,
                     seed=0,
                     popsize=10000).fit(X, y, test_set=X)
    t2 = time.time()
    print(t2 - t)
    assert t2 - t < 1
    assert gp._time_limit == 0.9


def test_add_repeated_args():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add, Min, Max
    y = cl.copy()
    for ff in [Add, Min, Max]:
        ff.nargs = 10
        gp = EvoDAG(generations=np.inf,
                    tournament_size=2,
                    early_stopping_rounds=100,
                    time_limit=0.9,
                    # multiple_outputs=True,
                    classifier=False,
                    all_inputs=True,
                    function_set=[ff],
                    pr_variable=1,
                    seed=0,
                    popsize=10000)
        gp.X = X
        # gp.nclasses(y)
        gp.y = y
        gp.create_population()
        print(gp.population.population)
        node = gp.random_offspring()
        print(node, node._variable, X.shape)
        assert len(node._variable) <= X.shape[1]
        ff.nargs = 2


def test_classification_mo():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=10,
                time_limit=0.9,
                multiple_outputs=True,
                all_inputs=True,
                seed=0,
                popsize=10000)
    gp.X = X
    gp.nclasses(y)
    y = gp._bagging_fitness.transform_to_mo(y)
    gp.y = [SparseArray.fromlist(x) for x in y.T]
    assert isinstance(gp._mask, list)
    gp.create_population()


def test_classification_mo2():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf, tournament_size=2,
                early_stopping_rounds=10, time_limit=0.9,
                multiple_outputs=True, all_inputs=True,
                remove_raw_inputs=False,
                seed=0, popsize=10000)
    gp.X = X
    gp.nclasses(y)
    y = gp._bagging_fitness.transform_to_mo(y)
    y = [SparseArray.fromlist(x) for x in y.T]
    gp = EvoDAG(generations=np.inf, tournament_size=2,
                early_stopping_rounds=10, time_limit=0.9,
                multiple_outputs=True, all_inputs=True, seed=0,
                remove_raw_inputs=False,
                popsize=10000).fit(X, y)
    m = gp.model()
    print([(x, x._variable, x.height) for x in m._hist])
    # assert False
    assert len(m.decision_function(gp.X)) == 3


def test_function_selection():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add, Min, Max
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                function_set=[Add, Min, Max],
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=100).fit(X, y, test_set=X)
    assert gp._function_selection_ins
    for i in range(gp._function_selection_ins.nfunctions):
        assert gp._function_selection_ins.avg_fitness(i) != 0


def test_init():
    from EvoDAG import EvoDAG
    m = EvoDAG.init().fit(X, cl)
    assert m._function_set[0].nargs == 60
    model = m.model()
    hy = model.predict(X)
    print(model.decision_function)
    print([x.full_array() for x in model.decision_function(X)])
    print((cl == hy).mean())
    print(hy, cl)
    assert (cl == hy).mean() > 0.9


def test_init_regression():
    from EvoDAG import EvoDAG
    m = EvoDAG.init(seed=10, classifier=False).fit(X, cl)
    assert m._function_set[0].nargs == 60
    hy = m.predict(X)
    assert np.unique(hy).shape[0] > 3


def test_finite():
    from EvoDAG import EvoDAG
    evo = EvoDAG.init()
    evo._finite = False
    evo.fit(X, cl)
    m = evo.model()
    hy = m.predict(X)
    print((hy == cl).mean(), [x.full_array() for x in m.decision_function(np.array(X))])
    assert (hy == cl).mean() > 0.9


def test_init_evodag_extras():
    from EvoDAG import EvoDAG
    from test_command_line import default_nargs
    m = EvoDAG.init(seed=10, popsize=10,
                    early_stopping_rounds=10).fit(X, cl)
    assert m.popsize == 10
    default_nargs()


def test_init_evodag_params_dict():
    from EvoDAG import EvoDAG
    from test_command_line import default_nargs
    import os
    import json
    kw = os.path.join(os.path.dirname(__file__), '..',
                      'conf', 'default_parameters.json')
    with open(kw) as fpt:
        kw = json.loads(fpt.read())
    m = EvoDAG.init(params_fname=kw, seed=10, popsize=10,
                    early_stopping_rounds=10).fit(X, cl)
    assert m.popsize == 10
    default_nargs()


def test_X_list():
    from EvoDAG import EvoDAG
    from test_command_line import default_nargs
    m = EvoDAG.init(seed=10, popsize=10,
                    early_stopping_rounds=10).fit(X.tolist(), cl)
    assert m.popsize == 10
    default_nargs()
    print(X.shape, len(m.X))
    assert len(m.X) == 4


def test_test_set_shuffle():
    from EvoDAG import EvoDAG
    from test_command_line import default_nargs
    m = EvoDAG.init(seed=10, popsize=10,
                    early_stopping_rounds=10).fit(X.tolist(), cl, test_set='shuffle')
    assert m.popsize == 10
    default_nargs()
    print(X.shape, len(m.X))
    assert len(m.X) == 4
    v = m.population.hist[-1]
    assert v.hy_test and len(v.hy_test) == np.unique(cl).shape[0]


def test_bug_naive_bayes():
    from EvoDAG import EvoDAG
    from test_command_line import default_nargs
    Xt = X.copy()
    y = cl.copy()
    mask = np.zeros(y.shape[0], dtype=np.bool)
    mask[y == 0] = True
    mask[y == 1] = True
    mask[np.where(y == 2)[0][:2]] = True
    m = EvoDAG.init(seed=11, popsize=10, fitness_function='macro-F1',
                    early_stopping_rounds=10).fit(Xt[mask], y[mask])
    assert m.popsize == 10
    default_nargs()
    print(X.shape, len(m.X))
    assert len(m.X) == 4


def test_orthogonal_selection():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add
    Xt = X.copy()
    y = cl.copy()
    m = EvoDAG.init(seed=11, popsize=10, orthogonal_selection=True,
                    fitness_function='macro-F1',
                    early_stopping_rounds=10).fit(Xt, y)
    assert m._orthogonal_selection
    args = m.get_args_orthogonal(Add)
    assert len(args)

def test_orthogonal_dot_selection():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add
    Xt = X.copy()
    y = cl.copy()
    m = EvoDAG.init(seed=11, popsize=10, orthogonal_dot_selection=True, 
                    fitness_function='macro-F1',
                    early_stopping_rounds=10).fit(Xt, y)
    assert m._orthogonal_dot_selection
    args = m.get_args_orthogonal_dot(Add)
    assert len(args)

def test_tojson():
    from EvoDAG import EvoDAG
    Xt = X.copy()
    y = cl.copy()
    m = EvoDAG.init(seed=0, popsize=10, orthogonal_selection=True,
                    fitness_function='macro-F1',
                    early_stopping_rounds=10).fit(Xt, y)
    model = m.model()
    print(model._hist)
    for i in model._hist:
        print(i.tojson())
    print(len(model._hist))


def test_orthogonal_selection_regression():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add
    Xt = X.copy()
    y = cl.copy()
    m = EvoDAG.init(seed=11, popsize=10, orthogonal_selection=True,
                    classifier=False, early_stopping_rounds=10).fit(Xt, y)
    assert m._orthogonal_selection
    args = m.get_args_orthogonal(Add)
    assert len(args)
    print(args)
    
def test_orthogonal_dot_selection_regression():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add
    Xt = X.copy()
    y = cl.copy()
    m = EvoDAG.init(seed=11, popsize=10, orthogonal_dot_selection=True,
                    classifier=False, early_stopping_rounds=10).fit(Xt, y)
    assert m._orthogonal_dot_selection
    args = m.get_args_orthogonal_dot(Add)
    assert len(args)
    print(args)


def test_negative_selection_false():
    from EvoDAG import EvoDAG
    from EvoDAG.population import SteadyState
    import numpy as np

    class P(SteadyState):
        def random_selection(self, negative=False):
            if negative:
                self._llamo = True
            return np.random.randint(self.popsize)
    
    Xt = X.copy()
    y = cl.copy()
    m = EvoDAG.init(seed=11, popsize=10, orthogonal_selection=True,
                    negative_selection=False, population_class=P,
                    classifier=False, early_stopping_rounds=10).fit(Xt, y)
    assert not m._negative_selection
    assert not m._p._negative_selection
    assert m._p._llamo
    m = EvoDAG.init(seed=11, popsize=10, orthogonal_selection=True,
                    negative_selection=True, population_class=P,
                    classifier=False, early_stopping_rounds=10).fit(Xt, y)
    try:
        m._p._llamo
    except AttributeError:
        return
    assert False
    



# Copyright 2016 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from test_root import X, cl
from nose.tools import assert_almost_equals
from test_command_line import default_nargs


def test_generational_generation():
    from EvoDAG.population import Generational
    from EvoDAG import EvoDAG
    function_set = [x for x in EvoDAG()._function_set if x.regression]
    gp = EvoDAG(population_class='Generational', classifier=False,
                function_set=function_set,
                popsize=10)
    gp.X = X
    y = cl.copy()
    y[y != 1] = -1
    gp.y = y
    gp.create_population()
    assert isinstance(gp.population, Generational)
    p = []
    for i in range(gp.popsize-1):
        a = gp.random_offspring()
        p.append(a)
        gp.replace(a)
    assert len(gp.population._inner) == (gp.popsize - 1)
    a = gp.random_offspring()
    p.append(a)
    gp.replace(a)
    assert len(gp.population._inner) == 0
    for a, b in zip(gp.population.population, p):
        assert a == b


def test_all_inputs():
    from EvoDAG import EvoDAG
    y = cl.copy()
    y[y != 1] = -1
    for pc in ['Generational', 'SteadyState']:
        gp = EvoDAG(population_class=pc,
                    all_inputs=True, classifier=False,
                    popsize=10)
        gp.X = X
        gp.y = y
        gp.create_population()
        assert len(gp.population.population) < 10
        for i in range(gp.population.popsize,
                       gp.population._popsize):
            a = gp.random_offspring()
            gp.replace(a)
        assert len(gp.population.population) == 10


def test_all_inputs2():
    from EvoDAG import EvoDAG
    y = cl.copy()
    y[y != 1] = -1
    gp = EvoDAG(population_class='Generational',
                all_inputs=True, classifier=False,
                pr_variable=1, popsize=3)
    gp.X = X
    gp.y = y
    gp.create_population()
    print(len(gp.population.population), len(gp.X))
    assert len(gp.population.population) == len(gp.X)
    for i in range(gp.popsize):
        a = gp.random_offspring()
        gp.replace(a)
    assert len(gp.population.population) == gp.popsize


def test_all_inputs3():
    from EvoDAG import EvoDAG
    y = cl.copy()
    y[y != 1] = -1
    gp = EvoDAG(population_class='SteadyState',
                all_inputs=True, classifier=False,
                pr_variable=1, popsize=3)
    gp.X = X
    gp.y = y
    gp.create_population()
    print(len(gp.population.population), len(gp.X))
    assert len(gp.population.population) == gp.population.popsize
    for i in range(gp.popsize):
        a = gp.random_offspring()
        gp.replace(a)
    assert len(gp.population.population) == gp.popsize
    

def test_all_init_popsize():
    from EvoDAG import EvoDAG
    y = cl.copy()
    y[y != 1] = -1
    gp = EvoDAG(population_class='Generational',
                all_inputs=True, classifier=False,
                early_stopping_rounds=1, pr_variable=1, popsize=2)
    gp.X = X
    gp.y = y
    gp.create_population()
    assert gp.init_popsize == len(gp.X)
    gp = EvoDAG(population_class='Generational', classifier=False,
                # all_inputs=True,
                early_stopping_rounds=1,
                popsize=2)
    gp.X = X
    gp.y = y
    gp.create_population()
    assert gp.init_popsize == gp.popsize


def test_random_generations():
    from EvoDAG import EvoDAG
    from EvoDAG.population import SteadyState

    class P(SteadyState):
        def random_selection(self, negative=False):
            raise RuntimeError('!')
    y = cl.copy()
    y[y != 1] = -1
    for pop in ['SteadyState', 'Generational', P]:
        gp = EvoDAG(population_class=pop, classifier=False,
                    all_inputs=True, random_generations=1,
                    early_stopping_rounds=1, popsize=2)
        gp.X = X
        gp.y = y
        gp.create_population()
        print(gp.population._random_generations)
        assert gp.population._random_generations == 1
        if pop == P:
            try:
                ind = gp.random_offspring()
                gp.replace(ind)
                assert False
            except RuntimeError:
                pass
        else:
            for i in range(3):
                gp.replace(gp.random_offspring())
            assert gp.population.generation == 2


def test_SteadyState_generation():
    from EvoDAG import EvoDAG
    y = cl.copy()
    y[y != 1] = -1
    gp = EvoDAG(population_class='SteadyState',
                all_inputs=True, classifier=False,
                early_stopping_rounds=1,
                popsize=2)
    gp.X = X
    gp.y = y
    gp.create_population()
    for i in range(3):
        gp.replace(gp.random_offspring())
    assert gp.population.generation == 2
    

def test_clean():
    from EvoDAG.node import Centroid
    from EvoDAG import EvoDAG
    Centroid.nargs = 0
    y = cl.copy()
    y[y != 1] = -1
    for pc in ['Generational', 'SteadyState']:
        gp = EvoDAG(population_class=pc, classifier=False,
                    popsize=5)
        gp.X = X
        gp.y = y
        gp.create_population()
        for i in range(10):
            v = gp.random_offspring()
            gp.replace(v)
        pop = gp.population.population
        esi = gp.population.estopping
        for i in gp.population._hist:
            print(i == esi, i in pop, i, '-'*10, i.fitness)
            if i == esi:
                assert i.hy is not None
            elif i in pop:
                assert i.hy is not None
        assert gp.population.estopping.hy is not None
    Centroid.nargs = 2


def test_density():
    from EvoDAG import EvoDAG
    Xc = X.copy()
    Xc[0, 0] = 0
    Xc[1, 1] = 0
    y = cl.copy()
    y[y != 1] = -1
    for pc in ['Generational', 'SteadyState']:
        gp = EvoDAG(population_class=pc, classifier=False,
                    popsize=5)
        gp.X = Xc
        gp.y = y
        gp.create_population()
        d = sum([x.hy.density for x in gp.population.population]) / gp.popsize
        print(d, gp.population.density, 'pop')
        assert gp.population.density == d
        for _ in range(3):
            a = gp.random_offspring()
            print(a.hy.density)
            gp.replace(a)
            d = sum([x.hy.density for x in gp.population.population]) / gp.popsize
            print(d, gp.population.density, 'replace')
            print(gp.population.density, d, '==')
            assert_almost_equals(gp.population.density, d)


def test_share_inputs():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(classifier=True, multiple_outputs=True,
                popsize=5, share_inputs=True)
    gp.fit(X, y)
    assert gp._share_inputs


def test_model_nvar():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(classifier=True, multiple_outputs=True,
                popsize=5, share_inputs=True)
    gp.fit(X, y)
    assert gp._share_inputs
    m = gp.model()
    print(X.shape)
    assert m.nvar == X.shape[1]
    try:
        m.predict(X[:, :3])
        assert False
    except RuntimeError:
        pass


def test_selectNumbers():
    from EvoDAG.cython_utils import SelectNumbers
    s = SelectNumbers([x for x in range(100)])
    a = s.get(10)
    print(a, len(a))
    assert len(a) == 10
    a = s.get(91)
    assert len(a) == 90
    s.pos -= 10
    a = s.get(10)
    assert len(a) == 10
    assert s.pos == 100
    assert s.empty()
    s.pos = 0
    while not s.empty():
        s.get_one()
    s = SelectNumbers([])
    assert s.empty()


def test_inputs():
    from EvoDAG.population import Inputs
    from EvoDAG.cython_utils import SelectNumbers
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(classifier=True, multiple_outputs=True,
                popsize=5, share_inputs=True)
    gp.X = X
    gp.nclasses(y)
    gp.y = y
    inputs = Inputs(gp, SelectNumbers([x for x in range(len(gp.X))]))
    func = inputs._func
    for f in func:
        inputs._func = [f]
        inputs._nfunc = 1
        v = inputs.input()
        assert v is not None
        inputs = Inputs(gp, SelectNumbers([x for x in range(len(gp.X))]))


def test_multiple_variables():
    import numpy as np
    from EvoDAG.population import Inputs
    from EvoDAG.cython_utils import SelectNumbers
    from EvoDAG import EvoDAG
    from SparseArray import SparseArray
    y = cl.copy()
    gp = EvoDAG(classifier=True, multiple_outputs=True,
                popsize=5, share_inputs=True)
    gp.X = X
    gp.X[-1]._eval_tr = SparseArray.fromlist([0 for x in range(gp.X[-1].hy.size())])
    gp.nclasses(y)
    gp.y = y
    inputs = Inputs(gp, SelectNumbers([x for x in range(len(gp.X))]))
    inputs._func = [inputs._func[-1]]
    inputs._nfunc = 1
    v = inputs.input()
    assert v is not None
    mask = np.array(gp._mask[0].full_array(), dtype=np.bool)
    D = np.array([x.hy.full_array() for x in gp.X]).T
    b = np.array(gp._ytr[0].full_array())
    coef = np.linalg.lstsq(D[mask], b[mask])[0]
    for a, b in zip(coef, v.weight[0]):
        assert_almost_equals(a, b)


def test_inputs_func_argument():
    from EvoDAG import EvoDAG

    class Error:
        nargs = 2
        min_nargs = 2
        classification = True
        regression = True

        def __init__(self, *args, **kwargs):
            raise RuntimeError('aqui')
    y = cl.copy()
    gp = EvoDAG(classifier=True, multiple_outputs=True,
                pr_variable=0, input_functions=[Error], popsize=5, share_inputs=True)
    gp.X = X
    gp.nclasses(y)
    gp.y = y
    try:
        gp.create_population()
        assert False
    except RuntimeError:
        pass
    gp = EvoDAG(classifier=True, multiple_outputs=True,
                pr_variable=0, input_functions=['NaiveBayes', 'NaiveBayesMN',
                                                'MultipleVariables'],
                popsize=5, share_inputs=True).fit(X, y)
    assert gp
    try:
        EvoDAG(classifier=True, multiple_outputs=True,
               pr_variable=0, input_functions=['NaiveBayesXX', 'NaiveBayesMN',
                                               'MultipleVariables'],
               popsize=5, share_inputs=True).fit(X, y)
    except AttributeError:
        pass


def test_restrictions():
    from EvoDAG.population import Inputs
    from EvoDAG.cython_utils import SelectNumbers
    from EvoDAG import EvoDAG
    from SparseArray import SparseArray
    y = cl.copy()
    gp = EvoDAG(classifier=True, multiple_outputs=True,
                popsize=5, share_inputs=True)
    gp.X = X
    gp.X[-1]._eval_tr = SparseArray.fromlist([0 for x in range(gp.X[-1].hy.size())])
    gp.nclasses(y)
    gp.y = y
    for c in [True, False]:
        gp._classifier = c
        tag = 'classification' if c else 'regression'
        inputs = Inputs(gp, SelectNumbers([x for x in range(len(gp.X))]))
        for x in inputs._funcs:
            print(x, tag)
            assert getattr(x, tag)


def test_inputs_func_argument_regression():
    from EvoDAG import EvoDAG

    class Error:
        nargs = 2
        min_nargs = 2
        classification = True
        regression = True

        def __init__(self, *args, **kwargs):
            raise RuntimeError('aqui')
    y = cl.copy()
    y[y == 0] = -1
    y[y > -1] = 1
    gp = EvoDAG(classifier=False, multiple_outputs=False,
                pr_variable=0, input_functions=[Error],
                popsize=5, share_inputs=True)
    gp.X = X
    gp.nclasses(y)
    gp.y = y
    try:
        gp.create_population()
        assert False
    except RuntimeError:
        pass


def get_remote_data():
    import os
    import subprocess
    if not os.path.isfile('evodag.params'):
        subprocess.call(['curl', '-O', 'http://ws.ingeotec.mx/~mgraffg/evodag_data/evodag.params'])
    if not os.path.isfile('train.sp'):
        subprocess.call(['curl', '-O', 'http://ws.ingeotec.mx/~mgraffg/evodag_data/train.sp'])


def test_HGeneration():
    import json
    import gzip
    import pickle
    from EvoDAG.utils import RandomParameterSearch
    from EvoDAG import EvoDAG

    get_remote_data()
    params = json.loads(open('evodag.params').read())
    try:
        with gzip.open('train.sp') as fpt:
            X = pickle.load(fpt)
            y = pickle.load(fpt)
    except ValueError:
        return
    params['population_class'] = 'HGenerational'
    params['pr_variable'] = 0.1
    kw = RandomParameterSearch.process_params(params)
    gp = EvoDAG(**kw).fit(X, y)
    assert gp


def test_HGeneration_pr_variable():
    import json
    from EvoDAG.utils import RandomParameterSearch
    from EvoDAG import EvoDAG
    get_remote_data()
    params = json.loads(open('evodag.params').read())
    params['population_class'] = 'HGenerational'
    params['pr_variable'] = 1.0
    kw = RandomParameterSearch.process_params(params)
    y = cl.copy()
    try:
        EvoDAG(**kw).fit(X, y)
    except AssertionError:
        return
    assert False


def test_all_variables_inputs():
    from EvoDAG.population import Inputs
    from EvoDAG.cython_utils import SelectNumbers
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(classifier=True, multiple_outputs=True,
                use_all_vars_input_functions=True,
                popsize=5, share_inputs=True)
    gp.X = X
    gp.nclasses(y)
    gp.y = y
    inputs = Inputs(gp, SelectNumbers([x for x in range(len(gp.X))]))
    func = inputs._func
    print(func, gp.nvar)
    for f in func:
        v = inputs.all_variables()
        assert v is not None
        assert isinstance(v, f)
    assert inputs._all_variables_index == len(func)


def test_input_functions():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG.init(input_functions=["NaiveBayes", "NaiveBayesMN", "Centroid"],
                     Centroid=2, time_limit=5)
    input_functions = [x for x in gp._input_functions]
    gp.fit(X, y)
    default_nargs()
    for a, b in zip(input_functions, gp.population.hist[:3]):
        print(b, a)
        assert isinstance(b, a)


def test_popsize_nvar():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG.init(popsize='nvar', time_limit=5)
    print(X.shape)
    gp.fit(X, y)
    default_nargs()
    assert gp.population._popsize == (X.shape[1] + len(gp._input_functions))



# Copyright 2016 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
import os
from test_root import X, cl


def default_nargs():
    from EvoDAG.node import Add, Mul, Div, Min, Max, Atan2, Hypot
    from EvoDAG.node import Fabs, Exp, Sqrt, Sin, Cos, Log1p,\
        Sq, Acos, Asin, Atan,\
        Tan, Cosh, Sinh, Tanh, Acosh, Asinh, Atanh,\
        Expm1, Log, Log2, Log10, Lgamma, Sign, Ceil, Floor,\
        NaiveBayes, NaiveBayesMN
    for f in [Add, Mul, Div, Min, Max, Atan2, Hypot, NaiveBayes, NaiveBayesMN]:
        f.nargs = 2
    for f in [Fabs, Exp, Sqrt, Sin, Cos, Log1p,
              Sq, Acos, Asin, Atan,
              Tan, Cosh, Sinh, Tanh, Acosh, Asinh, Atanh,
              Expm1, Log, Log2, Log10, Lgamma, Sign,
              Ceil, Floor]:
        f.nargs = 1


def training_set():
    import tempfile
    fname = tempfile.mktemp()
    with open(fname, 'w') as fpt:
        for x, v in zip(X, cl):
            l = x.tolist()
            l.append(v)
            fpt.write(','.join(map(str, l)))
            fpt.write('\n')
    return fname


def test_optimize_parameters():
    import os
    from EvoDAG.command_line import params
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag.gz', '-p3', '-e2', '-r2', fname]
    params()
    assert os.path.isfile('cache.evodag.gz')
    os.unlink('cache.evodag.gz')
    os.unlink(fname)
    default_nargs()


def test_cpu_cores():
    import os
    from EvoDAG.command_line import params
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '-u2', '--parameters',
                'cache.evodag.gz', '-p3', '-e2', '-r', '2', fname]
    params()
    assert os.path.isfile('cache.evodag.gz')
    os.unlink('cache.evodag.gz')
    os.unlink(fname)
    default_nargs()


def test_word2id():
    import tempfile
    from EvoDAG.command_line import params
    fname = tempfile.mktemp()
    id2word = dict([[0, 'a'], [1, 'b'], [2, 'c']])
    with open(fname, 'w') as fpt:
        for x, v in zip(X, cl):
            l = x.tolist()
            l.append(id2word[v])
            fpt.write(','.join(map(str, l)))
            fpt.write('\n')
    sys.argv = ['EvoDAG', '-C', '-Pparams.gz', '-r2', '-e2', '-p3', fname]
    params()
    os.unlink(fname)
    os.unlink('params.gz')
    default_nargs()


def test_json():
    from EvoDAG.command_line import params
    import tempfile
    import json
    fname = tempfile.mktemp()
    with open(fname, 'w') as fpt:
        for x, y in zip(X, cl):
            a = {k: v for k, v in enumerate(x)}
            a['klass'] = int(y)
            a['num_terms'] = len(x)
            fpt.write(json.dumps(a) + '\n')
    print("termine con el json")
    sys.argv = ['EvoDAG', '-C', '-Poutput.evodag', '--json',
                '-e1', '-p3', '-r2', fname]
    params()
    os.unlink(fname)
    print(open('output.evodag').read())
    os.unlink('output.evodag')
    default_nargs()


def test_params():
    import os
    import gzip
    from EvoDAG.command_line import params
    import json
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag.gz', '-p3', '-e2', '-r', '2', fname]
    params()
    os.unlink(fname)
    assert os.path.isfile('cache.evodag.gz')
    with gzip.open('cache.evodag.gz', 'rb') as fpt:
        try:
            d = fpt.read()
            a = json.loads(str(d, encoding='utf-8'))
        except TypeError:
            a = json.loads(d)
    os.unlink('cache.evodag.gz')
    assert len(a) == len([x for x in a if 'fitness' in x])
    print(a)
    default_nargs()


def test_parameters_values():
    import os
    from EvoDAG.command_line import params
    import json
    fname = training_set()
    with open('p.conf', 'w') as fpt:
        fpt.write(json.dumps(dict(popsize=['x'], multiple_outputs=[True])))
    sys.argv = ['EvoDAG', '-C', '-Pcache.evodag.gz', '-p3', '-e2',
                '--parameters-values', 'p.conf',
                '-r', '2', fname]
    try:
        params()
        assert False
    except ValueError:
        pass
    os.unlink('p.conf')
    os.unlink(fname)
    default_nargs()


def test_train():
    import os
    from EvoDAG.command_line import params, CommandLineTrain
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '-Pcache.evodag.gz',
                '-p3', '-e2', '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '-Pcache.evodag.gz',
                '-n2',
                '--model', 'model.evodag',
                '--test_set', fname, fname]
    c = CommandLineTrain()
    c.parse_args()
    assert os.path.isfile(c.data.test_set)
    assert os.path.isfile('model.evodag')
    os.unlink(fname)
    os.unlink('cache.evodag.gz')
    os.unlink('model.evodag')
    default_nargs()


def test_train_default_params():
    import os
    from EvoDAG.command_line import CommandLineTrain
    fname = training_set()
    sys.argv = ['EvoDAG', '-n2', '-C',
                '--model', 'model.evodag', fname]
    c = CommandLineTrain()
    c.parse_args()
    assert os.path.isfile('model.evodag')
    os.unlink(fname)
    os.unlink('model.evodag')
    default_nargs()


def test_predict():
    import os
    from EvoDAG.command_line import params, train, predict
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag.gz', '-p3', '-e2', '-r', '2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag.gz',
                '-n2',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--model', 'model.evodag', fname]
    predict()
    os.unlink(fname)
    os.unlink('cache.evodag.gz')
    os.unlink('model.evodag')
    assert os.path.isfile('output.evodag')
    os.unlink('output.evodag')
    default_nargs()


def test_generational():
    import os
    from EvoDAG.command_line import CommandLineParams
    import gzip
    import json
    fname = training_set()

    class C(CommandLineParams):
        def parse_args(self):
            self.data = self.parser.parse_args()
            self.data.population_class = 'Generational'
            if hasattr(self.data, 'regressor') and self.data.regressor:
                self.data.classifier = False
            self.main()

    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag.gz', '-p3', '-e3',
                '-r', '2', fname]
    c = C()
    c.parse_args()
    with gzip.open('cache.evodag.gz') as fpt:
        data = fpt.read()
        try:
            a = json.loads(str(data, encoding='utf-8'))
        except TypeError:
            a = json.loads(data)
    a = a[0]
    assert 'population_class' in a
    assert a['population_class'] == 'Generational'
    os.unlink('cache.evodag.gz')
    print(a)
    default_nargs()


def test_all_inputs():
    import os
    import json
    from EvoDAG.command_line import CommandLineParams
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag', '-p3', '-e2',
                '-r', '2', fname]
    c = CommandLineParams()
    c.parse_args()
    with open('cache.evodag') as fpt:
        a = json.loads(fpt.read())[0]
    assert 'all_inputs' in a
    os.unlink('cache.evodag')
    default_nargs()


def test_time_limit():
    import os
    from EvoDAG.command_line import params, train
    import json
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag', '-p3', '-e2',
                '--time-limit', '10',
                '-r', '2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    os.unlink(fname)
    with open('cache.evodag') as fpt:
        a = json.loads(fpt.read())[0]
    assert 'time_limit' in a
    os.unlink('cache.evodag')
    assert os.path.isfile('model.evodag')
    os.unlink('model.evodag')
    default_nargs()


def test_word2id2():
    import tempfile
    import os
    from EvoDAG.command_line import CommandLineParams
    fname = tempfile.mktemp()
    id2word = dict([[0, 'a'], [1, 'b'], [2, 'c']])
    with open(fname, 'w') as fpt:
        for x, v in zip(X, cl):
            l = x.tolist()
            l.append(id2word[v])
            fpt.write(','.join(map(str, l)))
            fpt.write('\n')
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r', '2', fname]
    c = CommandLineParams()
    c.parse_args()
    print(len(c.word2id))
    assert len(c.word2id) == 0
    os.unlink('cache.evodag')
    os.unlink(fname)
    default_nargs()


def test_decision_function():
    import os
    from EvoDAG.command_line import params, train, predict
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r', '1', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--decision-function',
                '--model', 'model.evodag', fname]
    predict()
    os.unlink(fname)
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    os.unlink('output.evodag')
    default_nargs()


def test_random_generations():
    import os
    import json
    from EvoDAG.command_line import params
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag', '-p3', '-e2',
                # '--random-generations', '1',
                '-r', '2', fname]
    params()
    os.unlink(fname)
    with open('cache.evodag') as fpt:
        a = json.loads(fpt.read())[0]
    assert 'random_generations' in a
    os.unlink('cache.evodag')
    default_nargs()


def test_predict_cpu():
    import os
    from EvoDAG.command_line import params, train, predict
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--decision-function',
                '-u2',
                '--model', 'model.evodag', fname]
    predict()
    os.unlink(fname)
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    os.unlink('output.evodag')
    default_nargs()


def test_classifier_params():
    import os
    import json
    from EvoDAG.command_line import params
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r', '2', fname]
    params()
    os.unlink(fname)
    with open('cache.evodag') as fpt:
        a = json.loads(fpt.read())[0]
    assert 'classifier' in a
    assert a['classifier']
    os.unlink('cache.evodag')
    default_nargs()


def test_regressor_params():
    import os
    import json
    from EvoDAG.command_line import params
    fname = training_set()
    sys.argv = ['EvoDAG', '-R', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r', '2', fname]
    params()
    os.unlink(fname)
    with open('cache.evodag') as fpt:
        a = json.loads(fpt.read())[0]
    assert 'classifier' in a
    assert not a['classifier']
    assert a['popsize'] == 3
    assert a['early_stopping_rounds'] == 1
    os.unlink('cache.evodag')
    default_nargs()
    

def test_time():
    import os
    import json
    from EvoDAG.command_line import params
    fname = training_set()
    sys.argv = ['EvoDAG', '-R', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r', '2', fname]
    params()
    os.unlink(fname)
    with open('cache.evodag') as fpt:
        a = json.loads(fpt.read())[0]
    assert 'classifier' in a
    assert not a['classifier']
    assert a['popsize'] == 3
    assert a['early_stopping_rounds'] == 1
    print(a['_time'])
    assert a['_time'] >= 0.0
    os.unlink('cache.evodag')
    default_nargs()


def test_utils_graphviz():
    import os
    from EvoDAG.command_line import params, train, utils
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--decision-function',
                '-u2',
                '--model', 'model.evodag', fname]
    sys.argv = ['EvoDAG', '-G', '-oevodag_gv',
                'model.evodag']

    utils()
    os.unlink(fname)
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    for i in range(2):
        os.unlink('evodag_gv/evodag-%s' % i)
    os.rmdir('evodag_gv')
    default_nargs()


def test_json_gzip():
    from EvoDAG.command_line import params
    import gzip
    import tempfile
    import json
    fname = tempfile.mktemp() + '.gz'
    with gzip.open(fname, 'wb') as fpt:
        for x, y in zip(X, cl):
            a = {k: v for k, v in enumerate(x)}
            a['klass'] = int(y)
            a['num_terms'] = len(x)
            try:
                fpt.write(bytes(json.dumps(a) + '\n', encoding='utf-8'))
            except TypeError:
                fpt.write(json.dumps(a) + '\n')
    print("termine con el json")
    sys.argv = ['EvoDAG', '-C', '-Poutput.evodag', '--json',
                '-e1', '-p3', '-r2', fname]
    params()
    os.unlink(fname)
    print(open('output.evodag').read())
    os.unlink('output.evodag')
    default_nargs()


def test_multiple_outputs():
    from EvoDAG.command_line import params
    import gzip
    import tempfile
    import json
    fname = tempfile.mktemp() + '.gz'
    with gzip.open(fname, 'wb') as fpt:
        for x, y in zip(X, cl):
            a = {k: v for k, v in enumerate(x)}
            a['klass'] = int(y)
            a['num_terms'] = len(x)
            try:
                fpt.write(bytes(json.dumps(a) + '\n', encoding='utf-8'))
            except TypeError:
                fpt.write(json.dumps(a) + '\n')
    print("termine con el json")
    sys.argv = ['EvoDAG', '-C', '-Poutput.evodag', '--json',
                '-e1', '-p3', '-r2', fname]
    params()
    os.unlink(fname)
    d = json.loads(open('output.evodag').read())
    assert d[0]['multiple_outputs']
    os.unlink('output.evodag')
    default_nargs()


def mo_training_set():
    import tempfile
    import numpy as np
    klass = np.unique(cl)
    y = np.empty((cl.shape[0], klass.shape[0]))
    y.fill(-1)
    for i, k in enumerate(klass):
        mask = k == cl
        y[mask, i] = 1
    fname = tempfile.mktemp()
    with open(fname, 'w') as fpt:
        for x, v in zip(X, y):
            l = x.tolist()
            l += v.tolist()
            fpt.write(','.join(map(str, l)))
            fpt.write('\n')
    return fname


def test_number_multiple_outputs_classification():
    import os
    from EvoDAG.command_line import params, train, predict
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-C', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2', '--output-dim=3',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--decision-function',
                '--model', 'model.evodag', fname]
    predict()
    os.unlink(fname)
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    l = open('output.evodag').readline()
    os.unlink('output.evodag')
    default_nargs()
    assert len(l.split(',')) == 3


def test_number_multiple_outputs_regression():
    import os
    from EvoDAG.command_line import params, train, predict
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-R', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2', '--output-dim=3',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--decision-function',
                '--model', 'model.evodag', fname]
    predict()
    os.unlink(fname)
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    l = open('output.evodag').readline()
    os.unlink('output.evodag')
    default_nargs()
    assert len(l.split(',')) == 3


def test_utils_params_stats():
    from EvoDAG.command_line import params, utils
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-R', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r3', fname]
    params()
    sys.argv = ['EvoDAG', '-P', '-o', fname, 'cache.evodag']
    utils()
    with open(fname) as fpt:
        print(fpt.read())
    os.unlink(fname)
    os.unlink('cache.evodag')
    default_nargs()


def test_model_fitness_vs():
    import os
    from EvoDAG.command_line import params, train, utils
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-R', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2', '--output-dim=3',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--fitness', 'model.evodag']
    utils()
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    default_nargs()


def test_raw_outputs_regression():
    import os
    from EvoDAG.command_line import params, train, predict
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-R', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2', '--output-dim=3',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--raw-outputs',
                '--model', 'model.evodag', fname]
    predict()
    os.unlink(fname)
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    l = open('output.evodag').readline()
    os.unlink('output.evodag')
    default_nargs()
    print(len(l.split(',')))
    assert len(l.split(',')) == 6


def test_raw_outputs_classification():
    import os
    from EvoDAG.command_line import params, train, predict
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-C', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2', '--output-dim=3',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--raw-outputs',
                '--model', 'model.evodag', fname]
    predict()
    os.unlink(fname)
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    l = open('output.evodag').readline()
    os.unlink('output.evodag')
    default_nargs()
    assert len(l.split(',')) == 6


def test_nan():
    import numpy as np
    from EvoDAG.command_line import CommandLineParams
    c = CommandLineParams()
    assert np.isfinite(c.inputs.convert_label('NaN'))
    default_nargs()


def test_model_size():
    import os
    from EvoDAG.command_line import params, train, utils
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-R', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2', '--output-dim=3',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--size', 'model.evodag']
    utils()
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    default_nargs()


def test_model_height():
    import os
    from EvoDAG.command_line import params, train, utils
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-R', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2', '--output-dim=3',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--height', 'model.evodag']
    utils()
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    default_nargs()


def test_utils_graphviz_terminals():
    import os
    from EvoDAG.command_line import params, train, utils
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--decision-function',
                '-u2',
                '--model', 'model.evodag', fname]
    sys.argv = ['EvoDAG', '-G', '-oevodag_gv',
                '--remove-terminals',
                'model.evodag']

    utils()
    os.unlink(fname)
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    for i in range(2):
        os.unlink('evodag_gv/evodag-%s' % i)
    os.rmdir('evodag_gv')
    default_nargs()


def test_json_gzip_dependent_variable():
    from EvoDAG.command_line import params
    import gzip
    import tempfile
    import json
    import os
    os.environ['KLASS'] = 'dependent'
    fname = tempfile.mktemp() + '.gz'
    with gzip.open(fname, 'wb') as fpt:
        for x, y in zip(X, cl):
            a = {k: v for k, v in enumerate(x)}
            a['dependent'] = int(y)
            a['num_terms'] = len(x)
            try:
                fpt.write(bytes(json.dumps(a) + '\n', encoding='utf-8'))
            except TypeError:
                fpt.write(json.dumps(a) + '\n')
    print("termine con el json")
    sys.argv = ['EvoDAG', '-C', '-Poutput.evodag', '--json',
                '-e1', '-p3', '-r2', fname]
    params()
    os.unlink(fname)
    print(open('output.evodag').read())
    os.unlink('output.evodag')
    default_nargs()
    del os.environ['KLASS']


def test_model_used_inputs_number():
    import os
    from EvoDAG.command_line import params, train, utils
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-R', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2', '--output-dim=3',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--used-inputs-number', 'model.evodag']
    utils()
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    default_nargs()


# def test_model_min_size():
#     import os
#     from EvoDAG.command_line import params, train
#     import gzip
#     import pickle
#     fname = mo_training_set()
#     sys.argv = ['EvoDAG', '--output-dim=3',
#                 '-R', '--parameters',
#                 'cache.evodag', '-p3', '-e2',
#                 '-r2', fname]
#     params()
#     sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
#                 '-n2', '--output-dim=3',
#                 '--min-size=6',
#                 '--model', 'model.evodag',
#                 '--test', fname, fname]
#     train()
#     os.unlink('cache.evodag')
#     with gzip.open('model.evodag') as fpt:
#         m = pickle.load(fpt)
#         for x in m.models:
#             print(x.size)
#             assert x.size >= 6
#     os.unlink('model.evodag')
#     default_nargs()


def test_logging():
    from EvoDAG.command_line import params
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-R', '--parameters', 'cache.evodag',
                '--verbose=0', '-p3', '-e2', '-r2', fname]
    params()


def test_model_seed():
    import os
    from EvoDAG.command_line import params, train
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-R', '--parameters',
                'cache.evodag', '-p3', '-e2',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2', '--output-dim=3', '--seed=1',
                '--model', 'model.evodag',
                '--test', fname, fname]
    c = train(output=True)
    os.unlink('cache.evodag')
    default_nargs()
    assert c.data.seed == 1


def test_create_ensemble():
    import os
    from EvoDAG.command_line import params, train, utils
    from EvoDAG.model import Ensemble
    import pickle
    import gzip
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-R', '--parameters',
                'cache.evodag', '-p3', '-e2',
                '-r2', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n1', '--output-dim=3', '--seed=1',
                '--model', 'model.evodag.1',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n1', '--output-dim=3', '--seed=2',
                '--model', 'model.evodag.2',
                '--test', fname, fname]
    train()
    os.unlink('cache.evodag')
    sys.argv = ['EvoDAG', '--create-ensemble', '-omodel.evodag',
                '-n3', 'model.evodag.* model.evodag.1']
    utils()
    os.unlink('model.evodag.1')
    os.unlink('model.evodag.2')
    with gzip.open('model.evodag', 'r') as fpt:
        ens = pickle.load(fpt)
    assert isinstance(ens, Ensemble)
    os.unlink('model.evodag')
    default_nargs()


def test_input_type_constraint_R():
    from EvoDAG.command_line import params
    from EvoDAG.utils import PARAMS
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-R', '--parameters', 'cache.evodag',
                '--verbose=0', '-p3', '-e2', '-r2', fname]
    c = params(output=True)
    _params = PARAMS.copy()
    c.if_type_contraint(_params)
    print(_params['input_functions'], 'MultipleVariables')
    assert _params['input_functions'][0] == 'MultipleVariables'


def test_input_type_constraint_C():
    from EvoDAG.command_line import params
    from EvoDAG.utils import PARAMS
    fname = mo_training_set()
    sys.argv = ['EvoDAG', '--output-dim=3',
                '-C', '--parameters', 'cache.evodag',
                '--verbose=0', '-p3', '-e2', '-r2', fname]
    c = params(output=True)
    _params = PARAMS.copy()
    c.if_type_contraint(_params)
    print(_params['input_functions'], 'MultipleVariables')
    assert len(_params['input_functions']) == 4
    

def test_params_files():
    import os
    from EvoDAG.command_line import params
    import shutil
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '-Pcache', '--only-paramsfiles',
                '-r', '2', fname]
    params()
    if os.path.isdir('cache'):
        shutil.rmtree('cache')
        default_nargs()
        return
    assert False


def test_get_best_params_files():
    import numpy as np
    import gzip
    import pickle
    import os
    from EvoDAG.command_line import params, train, utils
    import shutil
    from glob import glob
    if os.path.isdir('cache'):
        shutil.rmtree('cache')
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '-Pcache', '--only-paramsfiles',
                '-r', '2', fname]
    params()
    for p in glob(os.path.join('cache', '*_params.json')):
        basename = p.split('_params.json')[0]
        for s in range(3):
            model = basename + '_%s.model' % s
            sys.argv = ['EvoDAG', '-s%s' % s, '-P%s' % p, '-m%s' % model, '-n1', fname]
            train()
    R = []
    for p in range(2):
        l = []
        for s in range(3):
            with gzip.open(os.path.join('cache', '%s_%s.model' % (p, s))) as fpt:
                m = pickle.load(fpt)
                l.append(m.fitness_vs * -1)
        R.append((p, l))
    print(R)
    m = np.array([np.median(x[1]) for x in R])
    m = m - np.min(m)
    param = ['%s_params.json' % k for k, x in enumerate(m) if x == 0]
    sys.argv = ['EvoDAG', '--best-params-file', 'cache']
    c = utils(output=True)
    print(c.best_params, param)
    assert c.best_params in param
    sys.argv = ['EvoDAG', '-u2', '--best-params-file', 'cache']
    c = utils(output=True)
    print(c.best_params, param, '*')
    assert c.best_params in param
    if os.path.isdir('cache'):
        shutil.rmtree('cache')
        default_nargs()


def test_json2():
    from EvoDAG.command_line import params
    import tempfile
    import json
    fname = tempfile.mktemp()
    with open(fname, 'w') as fpt:
        for x, y in zip(X, cl):
            a = {}
            a['vec'] = [[k, v] for k, v in enumerate(x)]
            a['klass'] = int(y)
            a['vecsize'] = len(x)
            fpt.write(json.dumps(a) + '\n')
    print("termine con el json")
    sys.argv = ['EvoDAG', '-C', '-Poutput.evodag', '--json',
                '-e1', '-p3', '-r2', fname]
    params()
    os.unlink(fname)
    print(open('output.evodag').read())
    os.unlink('output.evodag')
    default_nargs()


def test_json3():
    from EvoDAG.command_line import params
    import tempfile
    import json
    fname = tempfile.mktemp()
    with open(fname, 'w') as fpt:
        for x, y in zip(X, cl):
            a = {}
            a['vec'] = x.tolist()
            a['klass'] = int(y)
            fpt.write(json.dumps(a) + '\n')
    print("termine con el json")
    sys.argv = ['EvoDAG', '-C', '-Poutput.evodag', '--json',
                '-e1', '-p3', '-r2', fname]
    params()
    os.unlink(fname)
    print(open('output.evodag').read())
    os.unlink('output.evodag')
    default_nargs()


def test_train_inner_model():
    import os
    import gzip
    from EvoDAG.command_line import CommandLineTrain
    fname = training_set()
    if not os.path.isdir('model.evodag_dir'):
        os.mkdir('model.evodag_dir')
    with gzip.open(os.path.join('model.evodag_dir', '0.evodag'), 'w') as fpt:
        fpt.write(b'')
    sys.argv = ['EvoDAG', '-n2', '-C',
                '--model', 'model.evodag', fname]
    c = CommandLineTrain()
    c.parse_args()
    assert os.path.isfile('model.evodag')
    os.unlink(fname)
    os.unlink('model.evodag')
    default_nargs()
    

def training_set_str():
    import tempfile
    fname = tempfile.mktemp()
    Xs = X.tolist()
    h = ['cero', 'uno', 'dos']
    y = [h[x] for x in cl]
    Xs[0][0] = 'dato'
    with open(fname, 'w') as fpt:
        for x, v in zip(Xs, y):
            l = x
            l.append(v)
            fpt.write(','.join(map(str, l)))
            fpt.write('\n')
    return fname


def test_word_label_2id():
    from EvoDAG.command_line import CommandLineTrain, predict
    import gzip
    import pickle
    fname = training_set_str()
    sys.argv = ['EvoDAG', '-n2', '-C',
                '--model', 'model.evodag', fname]
    c = CommandLineTrain()
    c.parse_args()
    with gzip.open('model.evodag') as fpt:
        pickle.load(fpt)
        a = pickle.load(fpt)
        b = pickle.load(fpt)
    assert len(a) == 1
    assert 'dato' in a
    for k in b.keys():
        assert k in ['cero', 'uno', 'dos']
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--model', 'model.evodag', fname]
    predict()
    assert os.path.isfile('model.evodag')
    os.unlink(fname)
    os.unlink('model.evodag')
    default_nargs()
    with open('output.evodag') as fpt:
        l = fpt.readlines()
    os.unlink('output.evodag')
    for x in l:
        assert x.strip() in ['cero', 'uno', 'dos']


def test_train_kw():
    import os
    from EvoDAG.command_line import CommandLineTrain
    fname = training_set()
    sys.argv = ['EvoDAG', '-n2', '-C',
                '--kw', '{"popsize": 10, "early_stopping_rounds": 10}',
                '--model', 'model.evodag', fname]
    c = CommandLineTrain()
    c.parse_args()
    assert os.path.isfile('model.evodag')
    os.unlink(fname)
    os.unlink('model.evodag')
    default_nargs()


def test_train_fitness_function():
    import os
    from EvoDAG.command_line import CommandLineTrain
    fname = training_set()
    sys.argv = ['EvoDAG', '-n2', '-C',
                '--kw', '{"fitness_function": "F1", "F1_index": 0}',
                '--model', 'model.evodag', fname]
    c = CommandLineTrain()
    c.parse_args()
    assert os.path.isfile('model.evodag')
    os.unlink(fname)
    os.unlink('model.evodag')
    default_nargs()


def test_train_function_set():
    import os
    from EvoDAG.command_line import CommandLineTrain
    from EvoDAG.node import Add
    fname = training_set()
    sys.argv = ['EvoDAG', '-n2', '-C',
                '--kw', '{"popsize": 10, "early_stopping_rounds": 10, "Add": 10}',
                '--model', 'model.evodag', fname]
    c = CommandLineTrain()
    c.parse_args()
    assert os.path.isfile('model.evodag')
    os.unlink(fname)
    os.unlink('model.evodag')
    assert Add.nargs == 10
    default_nargs()


def test_predict_proba():
    import os
    from EvoDAG.command_line import params, train, predict
    fname = training_set()
    sys.argv = ['EvoDAG', '-C', '--parameters',
                'cache.evodag', '-p3', '-e1',
                '-r', '1', fname]
    params()
    sys.argv = ['EvoDAG', '--parameters', 'cache.evodag',
                '-n2',
                '--model', 'model.evodag',
                '--test', fname, fname]
    train()
    sys.argv = ['EvoDAG', '--output', 'output.evodag',
                '--predict-proba',
                '--model', 'model.evodag', fname]
    predict()
    os.unlink(fname)
    os.unlink('cache.evodag')
    os.unlink('model.evodag')
    os.unlink('output.evodag')
    default_nargs()



# Copyright 2016 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np


def test_parameter_grid_results():
    from EvoDAG.utils import RandomParameterSearch
    a = RandomParameterSearch()
    assert len([x for x in a]) == 1468


def test_process_params():
    from EvoDAG.utils import RandomParameterSearch
    from EvoDAG import EvoDAG
    rs = RandomParameterSearch(npoints=1)
    args = [x for x in rs][0]
    evo = EvoDAG(**rs.process_params(args))
    params = evo.get_params()
    for k, v in args.items():
        if k in params:
            if k == 'generations':
                v = np.inf
            print(k, v, params[k])
            if isinstance(v, list):
                for a, b in zip(v, params[k]):
                    assert a == b.__name__
            elif hasattr(params[k], '__name__'):
                assert v == params[k].__name__
            else:
                assert v == params[k]


def test_random_parameter_search_len():
    from itertools import product
    from EvoDAG.utils import RandomParameterSearch
    params = {'a': [1, 2, 3],
              'b': [-1, -2],
              'c': ['a', 'b', 'c', 'd', 'e']}
    rs = RandomParameterSearch(params=params)
    params = sorted(params.items())
    a = product(*[x[1] for x in params])
    assert len([x for x in a]) == len(rs)


def test_random_parameter_search_getitem():
    from itertools import product
    from EvoDAG.utils import RandomParameterSearch
    params = {'a': [1, 2, 3],
              'b': [-1, -2],
              'c': ['a', 'b', 'c', 'd', 'e']}
    rs = RandomParameterSearch(params=params)
    params = sorted(params.items())
    a = product(*[x[1] for x in params])
    for k, v in enumerate(a):
        v1 = [x[1] for x in sorted(rs[k].items())]
        print(v, v1)
        assert np.all([x == y for x, y in zip(v, v1)])

    
def test_params():
    from EvoDAG.utils import RandomParameterSearch
    rs = RandomParameterSearch(npoints=734)
    l = np.unique([x['random_generations'] for x in rs])
    assert len(l) == 2
    l = np.unique([x['population_class'] for x in rs])
    assert len(l) == 2


def test_iter_params():
    from EvoDAG.utils import RandomParameterSearch
    params = dict(popsize=[100, 1000],
                  population_class=['SteadyState'],
                  early_stopping_rounds=[50, 1000])
    rs = RandomParameterSearch(params=params, npoints=5)
    args = [x for x in rs]
    assert len(args) == 4


def test_constraints():
    from EvoDAG.utils import RandomParameterSearch
    params = dict(popsize=[100, 1000],
                  population_class=['Generational',
                                    'SteadyState'],
                  early_stopping_rounds=[50, 1000])
    rs = RandomParameterSearch(params=params, npoints=32)
    args = [x for x in rs if x['population_class'] == 'Generational']
    assert len(args) == 2


def test_constraints2():
    from EvoDAG.utils import RandomParameterSearch
    params = dict(popsize=[100, 1000],
                  population_class=['Generational',
                                    'SteadyState'],
                  early_stopping_rounds=[100, 1000])
    rs = RandomParameterSearch(params=params, npoints=32, training_size=70)
    args = [x for x in rs if x['population_class'] == 'Generational']
    assert len(args) == 1


def test_popsize_constraint():
    from EvoDAG.utils import RandomParameterSearch
    params = dict(popsize=[100, 1000],
                  population_class=['Generational',
                                    'SteadyState'],
                  early_stopping_rounds=[100, 1000])
    rs = RandomParameterSearch(params=params, npoints=32, training_size=70)
    popsize = [x['popsize'] for x in rs]
    r = [x for x in popsize if x > 70]
    assert len(r) == 0


def test_inputs_vec():
    from test_root import X, cl
    from EvoDAG.utils import Inputs
    D = []
    for x, y in zip(X, cl):
        a = {}
        a['vec'] = x.tolist()
        a['klass'] = int(y)
        D.append(a)
    inputs = Inputs()
    X, y = inputs.read_data_json_vec(None, D)


def test_inputs_vecsize():
    from test_root import X, cl
    from EvoDAG.utils import Inputs
    D = []
    for x, y in zip(X, cl):
        a = {}
        a['vec'] = [[k, v] for k, v in enumerate(x)]
        a['klass'] = int(y)
        a['vecsize'] = len(x)
        D.append(a)
    inputs = Inputs()
    X, y = inputs.read_data_json_vec(None, D)


def test_inputs():
    from test_root import X, cl
    from EvoDAG.utils import Inputs
    D = []
    for x, y in zip(X, cl):
        a = {k: v for k, v in enumerate(x)}
        a['klass'] = int(y)
        a['num_terms'] = len(x)
        D.append(a)
    inputs = Inputs()
    X, y = inputs.read_data_json(None, D)
    



# Copyright 2017 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from EvoDAG.function_selection import FunctionSelection
from nose.tools import assert_almost_equals


def test_seed():
    a = FunctionSelection(seed=0, nfunctions=10)
    rf = a.random_function()
    flag = False
    for seed in range(1, 10):
        b = FunctionSelection(seed=seed, nfunctions=10)
        if b.random_function() != rf:
            flag = True
    assert flag
    for seed in range(1, 10):
        b = FunctionSelection(seed=0, nfunctions=10)
        if b.random_function() != rf:
            assert False


def test_setitem():
    b = FunctionSelection(seed=0, nfunctions=10)
    assert b.fitness[3] == 0
    assert b.times[3] == 0
    b[3] = 3.2
    assert b.fitness[3] == 3.2
    assert b.times[3] == 1


def test_avg_fitness():
    b = FunctionSelection(seed=0, nfunctions=10)
    b[3] = 3.2
    b[3] = 5
    print(b.avg_fitness(3), (3.2 + 5) / 2.)
    assert b.avg_fitness(3) == (3.2 + 5) / 2.


def test_tournament():
    b = FunctionSelection(seed=0, nfunctions=10)
    for i in range(100):
        c = b.tournament()
        b[c] = -3.2
    for i in range(10):
        print(b.avg_fitness(i), -3.2)
        assert_almost_equals(b.avg_fitness(i), -3.2)


def test_number_args():
    for s in range(10):
        a = FunctionSelection(seed=s, nfunctions=2, nargs=[1, 2])
        index = a.tournament()
        assert index == 1


def test_density():
    a = FunctionSelection(seed=0, nfunctions=3, nargs=[1, 2, 2], density_safe=[1, 2])
    assert a.min_density == 0 and a.density == 1.0
    a.min_density = 0.9
    a.density = 0.7
    for i in range(10):
        index = a.tournament()
        assert index == 1 or index == 2
    a = FunctionSelection(seed=0, nfunctions=2, nargs=[1, 2], density_safe=[1])
    assert a.min_density == 0 and a.density == 1.0
    a.min_density = 0.9
    a.density = 0.7
    index = a.tournament()
    assert index == 1


def test_unfeasible_functions():
    a = FunctionSelection(seed=0, nfunctions=3, nargs=[1, 2, 2], density_safe=[1, 2])
    assert len(a.unfeasible_functions) == 0
    a.unfeasible_functions.add(1)
    a.unfeasible_functions.add(0)
    x = a.tournament()
    assert x == 2
    a = FunctionSelection(seed=0, nfunctions=3, nargs=[1, 2, 2])
    assert len(a.unfeasible_functions) == 0
    a.unfeasible_functions.add(1)
    a.unfeasible_functions.add(0)
    x = a.tournament()
    assert x == 2
    a.unfeasible_functions.add(2)
    x = a.tournament()
